from jury.metrics._core import (
    AutoMetric,
    EvaluationInstance,
    LanguageGenerationInstance,
    Metric,
    MetricForLanguageGeneration,
    MetricForSequenceClassification,
    MetricForSequenceLabeling,
    MetricForTask,
    SequenceClassificationInstance,
    SequenceLabelingInstance,
    list_metrics,
    load_metric,
)
from jury.metrics.accuracy import Accuracy
from jury.metrics.bartscore import Bartscore
from jury.metrics.bertscore import Bertscore
from jury.metrics.bleu import Bleu
from jury.metrics.bleurt import Bleurt
from jury.metrics.cer import CER
from jury.metrics.chrf import CHRF
from jury.metrics.comet import Comet
from jury.metrics.f1 import F1
from jury.metrics.meteor import Meteor
from jury.metrics.precision import Precision
from jury.metrics.prism import Prism
from jury.metrics.recall import Recall
from jury.metrics.rouge import Rouge
from jury.metrics.sacrebleu import Sacrebleu
from jury.metrics.squad import Squad
from jury.metrics.ter import TER
from jury.metrics.wer import WER



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
ChrF(++) (Character n-gram F-score) metric. The part of this file is adapted from HuggingFace's
datasets package implementation of CHRF metric. See
https://github.com/huggingface/datasets/blob/master/metrics/chrf/chrf.py
"""
from typing import Callable, Dict, List, Tuple, Union

import datasets
from packaging import version

from jury.collator import Collator
from jury.metrics import LanguageGenerationInstance, MetricForLanguageGeneration
from jury.metrics._core.utils import PackagePlaceholder, requirement_message

# `import sacrebleu as scb` placeholder
scb = PackagePlaceholder(version="2.0.0")


_CITATION = """\
@inproceedings{popovic-2015-chrf,
    title = "chr{F}: character n-gram {F}-score for automatic {MT} evaluation",
    author = "Popovi{\'c}, Maja",
    booktitle = "Proceedings of the Tenth Workshop on Statistical Machine Translation",
    month = sep,
    year = "2015",
    address = "Lisbon, Portugal",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/W15-3049",
    doi = "10.18653/v1/W15-3049",
    pages = "392--395",
}
@inproceedings{popovic-2017-chrf,
    title = "chr{F}++: words helping character n-grams",
    author = "Popovi{\'c}, Maja",
    booktitle = "Proceedings of the Second Conference on Machine Translation",
    month = sep,
    year = "2017",
    address = "Copenhagen, Denmark",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/W17-4770",
    doi = "10.18653/v1/W17-4770",
    pages = "612--618",
}
@inproceedings{post-2018-call,
    title = "A Call for Clarity in Reporting {BLEU} Scores",
    author = "Post, Matt",
    booktitle = "Proceedings of the Third Conference on Machine Translation: Research Papers",
    month = oct,
    year = "2018",
    address = "Belgium, Brussels",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/W18-6319",
    pages = "186--191",
}
"""

_DESCRIPTION = """\
ChrF and ChrF++ are two MT evaluation metrics. They both use the F-score statistic for character n-gram matches,
and ChrF++ adds word n-grams as well which correlates more strongly with direct assessment. We use the implementation
that is already present in sacrebleu.
The implementation here is slightly different from sacrebleu in terms of the required input format. The length of
the references and hypotheses lists need to be the same, so you may need to transpose your references compared to
sacrebleu's required input format. See https://github.com/huggingface/datasets/issues/3154#issuecomment-950746534
See the README.md file at https://github.com/mjpost/sacreBLEU#chrf--chrf for more information.
"""

_KWARGS_DESCRIPTION = """
Produces ChrF(++) scores for hypotheses given reference translations.
Args:
    predictions: The system stream (a sequence of segments).
    references: A list of one or more reference streams (each a sequence of segments).
    char_order: Character n-gram order.
    word_order: Word n-gram order. If equals to 2, the metric is referred to as chrF++.
    beta: Determine the importance of recall w.r.t precision.
    lowercase: Enable case-insensitivity.
    whitespace: If `True`, include whitespaces when extracting character n-grams.
    eps_smoothing: If `True`, applies epsilon smoothing similar
    to reference chrF++.py, NLTK and Moses implementations. Otherwise,
    it takes into account effective match order similar to sacreBLEU < 2.0.0.
Returns:
    'score': The chrF (chrF++) score,
    'char_order': The character n-gram order,
    'word_order': The word n-gram order. If equals to 2, the metric is referred to as chrF++,
    'beta': Determine the importance of recall w.r.t precision
Examples:
    >>> predictions = [["the cat is on the mat", "There is cat playing on the mat"], ["Look! a wonderful day."]]
    >>> references = [
        ["the cat is playing on the mat.", "The cat plays on the mat."], 
        ["Today is a wonderful day", "The weather outside is wonderful."]
    ]
    >>> chrf = jury.load_metric("chrf")
    >>> results = chrf.compute(predictions=prediction, references=reference)
    >>> print(results)
    {'chrf': {'score': 0.29778203723986857, 'char_order': 6, 'word_order': 0, 'beta': 2}}
"""


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class CHRFForLanguageGeneration(MetricForLanguageGeneration):
    def _info(self):
        if version.parse(scb.__version__) < version.parse("1.4.12"):
            raise ImportWarning(
                "To use `sacrebleu`, the module `sacrebleu>=1.4.12` is required, and the current version of "
                "`sacrebleu` doesn't match this condition.\nYou can install it with `pip install sacrebleu>=1.4.12`."
            )
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            homepage="https://github.com/mjpost/sacreBLEU#chrf--chrf",
            inputs_description=_KWARGS_DESCRIPTION,
            features=self._default_features,
            codebase_urls=["https://github.com/mjpost/sacreBLEU#chrf--chrf"],
            reference_urls=[
                "https://github.com/m-popovic/chrF",
            ],
        )

    def _download_and_prepare(self, dl_manager):
        global scb
        global CHRFScorer
        try:
            from sacrebleu import CHRF as CHRFScorer
        except ModuleNotFoundError:
            raise ModuleNotFoundError(requirement_message(path="chrf", package_name="sacrebleu"))
        else:
            super(CHRFForLanguageGeneration, self)._download_and_prepare(dl_manager)

    def _validate_references(self, references: Collator) -> None:
        references_per_prediction = len(references[0])
        if any(len(refs) != references_per_prediction for refs in references):
            raise ValueError("Sacrebleu requires the same number of references for each prediction")

    def _compute_chrf_score(
        self, predictions: Union[str, List[str]], references: Union[str, List[str]], **kwargs
    ) -> Tuple[float, int, int, int]:
        if kwargs.get("char_order") is None:
            kwargs["char_order"] = CHRFScorer.CHAR_ORDER
        if kwargs.get("word_order") is None:
            kwargs["word_order"] = CHRFScorer.WORD_ORDER
        if kwargs.get("beta") is None:
            kwargs["beta"] = CHRFScorer.BETA
        sb_chrf = CHRFScorer(**kwargs)
        output = sb_chrf.corpus_score(predictions, references)

        return (
            output.score / 100,
            output.char_order,
            output.word_order,
            output.beta,
        )

    def _compute_single_pred_single_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        char_order: int = None,
        word_order: int = None,
        beta: int = None,
        lowercase: bool = False,
        whitespace: bool = False,
        eps_smoothing: bool = False,
    ):
        score, c_ord, w_ord, beta = self._compute_chrf_score(
            predictions,
            references,
            char_order=char_order,
            word_order=word_order,
            beta=beta,
            lowercase=lowercase,
            whitespace=whitespace,
            eps_smoothing=eps_smoothing,
        )
        return {"score": score, "char_order": c_ord, "word_order": w_ord, "beta": beta}

    def _compute_single_pred_multi_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        char_order: int = None,
        word_order: int = None,
        beta: int = None,
        lowercase: bool = False,
        whitespace: bool = False,
        eps_smoothing: bool = False,
    ):
        # SacreBleu inherently supports multiple references.
        return self._compute_single_pred_single_ref(
            predictions=predictions,
            references=references,
            reduce_fn=reduce_fn,
            char_order=char_order,
            word_order=word_order,
            beta=beta,
            lowercase=lowercase,
            whitespace=whitespace,
            eps_smoothing=eps_smoothing,
        )

    def _compute_multi_pred_multi_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        char_order: int = None,
        word_order: int = None,
        beta: int = None,
        lowercase: bool = False,
        whitespace: bool = False,
        eps_smoothing: bool = False,
    ):
        scores = []
        for preds, refs in zip(predictions, references):
            pred_scores = []
            for pred in preds:
                score, _, _, _ = self._compute_chrf_score(
                    pred,
                    refs,
                    char_order=char_order,
                    word_order=word_order,
                    beta=beta,
                    lowercase=lowercase,
                    whitespace=whitespace,
                    eps_smoothing=eps_smoothing,
                )
                pred_scores.append(score)
            scores.append(float(reduce_fn(pred_scores)))

        return {
            "score": sum(scores) / len(scores),
            "char_order": char_order or CHRFScorer.CHAR_ORDER,
            "word_order": word_order or CHRFScorer.WORD_ORDER,
            "beta": beta or CHRFScorer.BETA,
        }

    def evaluate(
        self, predictions: Collator, references: Collator, reduce_fn: Callable = None, **kwargs
    ) -> Dict[str, float]:
        if predictions.can_collapse() and references.can_collapse():
            predictions = predictions.collapse()
            eval_fn = self._compute_single_pred_single_ref
        elif predictions.can_collapse() and not references.can_collapse():
            predictions = predictions.collapse()
            eval_fn = self._compute_single_pred_multi_ref
        else:
            eval_fn = self._compute_multi_pred_multi_ref
        self._validate_references(references)
        return eval_fn(predictions=predictions, references=references, reduce_fn=reduce_fn, **kwargs)



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
METEOR metric. The part of this file is adapted from HuggingFace's
datasets package implementation of METEOR metric. See
https://github.com/huggingface/datasets/blob/master/metrics/meteor/meteor.py
"""

from typing import Callable, Dict, Tuple

import datasets
import numpy as np
from nltk import __version__ as NLTK_VERSION
from nltk.translate import meteor_score
from packaging.version import Version

if Version(NLTK_VERSION) < Version("3.6.6"):
    raise EnvironmentError(
        f"Version constraints does not hold for 'nltk', expected version >=3.6.6, got {NLTK_VERSION}."
    )

from jury.collator import Collator
from jury.metrics._core import MetricForLanguageGeneration
from jury.tokenizer import DefaultTokenizer, TokenizerWrapper

_CITATION = """\
@inproceedings{banarjee2005,
  title     = {{METEOR}: An Automatic Metric for {MT} Evaluation with Improved Correlation with Human Judgments},
  author    = {Banerjee, Satanjeev  and Lavie, Alon},
  booktitle = {Proceedings of the {ACL} Workshop on Intrinsic and Extrinsic Evaluation Measures for Machine Translation and/or Summarization},
  month     = jun,
  year      = {2005},
  address   = {Ann Arbor, Michigan},
  publisher = {Association for Computational Linguistics},
  url       = {https://www.aclweb.org/anthology/W05-0909},
  pages     = {65--72},
}
"""

_DESCRIPTION = """\
METEOR, an automatic metric for machine translation evaluation
that is based on a generalized concept of unigram matching between the
machine-produced translation and human-produced reference translations.
Unigrams can be matched based on their surface forms, stemmed forms,
and meanings; furthermore, METEOR can be easily extended to include more
advanced matching strategies. Once all generalized unigram matches
between the two strings have been found, METEOR computes a score for
this matching using a combination of unigram-precision, unigram-recall, and
a measure of fragmentation that is designed to directly capture how
well-ordered the matched words in the machine translation are in relation
to the reference.

METEOR gets an R correlation value of 0.347 with human evaluation on the Arabic
data and 0.331 on the Chinese data. This is shown to be an improvement on
using simply unigram-precision, unigram-recall and their harmonic F1
combination.
"""

_KWARGS_DESCRIPTION = """
Computes METEOR score of translated segments against one or more references.
Args:
    predictions: list of predictions to score. Each prediction
        should be a string with tokens separated by spaces.
    references: list of reference for each prediction. Each
        reference should be a string with tokens separated by spaces.
    alpha: Parameter for controlling relative weights of precision and recall. default: 0.9
    beta: Parameter for controlling shape of penalty as a function of fragmentation. default: 3
    gamma: Relative weight assigned to fragmentation penalty. default: 0.5
Returns:
    'score': meteor score.
Examples:

    >>> meteor = jury.load_metric("meteor")
    >>> predictions = [["the cat is on the mat", "There is cat playing on the mat"], ["Look! a wonderful day."]]
    >>> references = [
        ["the cat is playing on the mat.", "The cat plays on the mat."], 
        ["Today is a wonderful day", "The weather outside is wonderful."]
    ]
    >>> results = meteor.compute(predictions=predictions, references=references)
    >>> print(results)
    {'meteor': {'score': 0.5420511682934044}}
"""


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class MeteorForLanguageGeneration(MetricForLanguageGeneration):
    def __init__(self, resulting_name: str = None, compute_kwargs: Dict = None, **kwargs):
        self.should_change_resulting_name = resulting_name is None
        self.tokenizer = DefaultTokenizer()
        super().__init__(resulting_name=resulting_name, compute_kwargs=compute_kwargs, **kwargs)

    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=self._default_features,
            codebase_urls=["https://github.com/nltk/nltk/blob/develop/nltk/translate/meteor_score.py"],
            reference_urls=[
                "https://www.nltk.org/api/nltk.translate.html#module-nltk.translate.meteor_score",
                "https://en.wikipedia.org/wiki/METEOR",
            ],
        )

    def _download_and_prepare(self, dl_manager):
        import nltk

        nltk.download("wordnet", quiet=True)
        nltk.download("punkt", quiet=True)
        nltk.download("omw-1.4", quiet=True)

    def _preprocess(self, predictions: Collator, references: Collator) -> Tuple[Collator, Collator]:
        tokenizer_wrapper = TokenizerWrapper(self.tokenizer)
        return tokenizer_wrapper.tokenize(predictions, references)

    def _compute_single_pred_single_ref(
        self, predictions: Collator, references: Collator, reduce_fn: Callable = None, alpha=0.9, beta=3, gamma=0.5
    ):
        scores = [
            meteor_score.single_meteor_score(ref, pred, alpha=alpha, beta=beta, gamma=gamma)
            for ref, pred in zip(references, predictions)
        ]
        return {"score": self._reduce_scores(scores, reduce_fn=np.mean)}

    def _compute_single_pred_multi_ref(
        self, predictions: Collator, references: Collator, reduce_fn: Callable = None, alpha=0.9, beta=3, gamma=0.5
    ):
        scores = [
            meteor_score.meteor_score(references=ref, hypothesis=pred, alpha=alpha, beta=beta, gamma=gamma)
            for ref, pred in zip(references, predictions)
        ]
        return {"score": self._reduce_scores(scores, reduce_fn=np.mean)}

    def _compute_multi_pred_multi_ref(
        self, predictions: Collator, references: Collator, reduce_fn: Callable = None, alpha=0.9, beta=3, gamma=0.5
    ):
        scores = []
        for pred, ref in zip(predictions, references):
            score = [
                meteor_score.meteor_score(references=ref, hypothesis=p, alpha=alpha, beta=beta, gamma=gamma)
                for p in pred
            ]
            reduced_score = reduce_fn(score)
            scores.append(reduce_fn(reduced_score))
        return {"score": self._reduce_scores(scores, reduce_fn=np.mean)}



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
SeqEval metric. The part of this file is adapted from seqeval implementation
of datasets package. See
https://github.com/huggingface/datasets/blob/master/metrics/seqeval/seqeval.py
"""
import importlib
from typing import Dict, List, Optional, Union

import datasets

from jury.metrics._core import MetricForSequenceLabeling, SequenceLabelingInstance
from jury.metrics._core.utils import PackagePlaceholder, requirement_message

# `import seqeval` placeholder
seqeval = PackagePlaceholder(version="1.2.2")


_CITATION = """\
@inproceedings{ramshaw-marcus-1995-text,
    title = "Text Chunking using Transformation-Based Learning",
    author = "Ramshaw, Lance  and
      Marcus, Mitch",
    booktitle = "Third Workshop on Very Large Corpora",
    year = "1995",
    url = "https://www.aclweb.org/anthology/W95-0107",
}
@misc{seqeval,
  title={{seqeval}: A Python framework for sequence labeling evaluation},
  url={https://github.com/chakki-works/seqeval},
  note={Software available from https://github.com/chakki-works/seqeval},
  author={Hiroki Nakayama},
  year={2018},
}
"""

_DESCRIPTION = """\
seqeval is a Python framework for sequence labeling evaluation.
seqeval can evaluate the performance of chunking tasks such as named-entity recognition, part-of-speech tagging, semantic role labeling and so on.

This is well-tested by using the Perl script conlleval, which can be used for
measuring the performance of a system that has processed the CoNLL-2000 shared task data.

seqeval supports following formats:
IOB1
IOB2
IOE1
IOE2
IOBES

See the [README.md] file at https://github.com/chakki-works/seqeval for more information.
"""

_KWARGS_DESCRIPTION = """
Produces labelling scores along with its sufficient statistics
from a source against one or more references.

Args:
    predictions: List of List of predicted labels (Estimated targets as returned by a tagger)
    references: List of List of reference labels (Ground truth (correct) target values)
    suffix: True if the IOB prefix is after type, False otherwise. default: False
    scheme: Specify target tagging scheme. Should be one of ["IOB1", "IOB2", "IOE1", "IOE2", "IOBES", "BILOU"].
        default: None
    mode: Whether to count correct entity labels with incorrect I/B tags as true positives or not.
        If you want to only count exact matches, pass mode="strict". default: None.
    sample_weight: Array-like of shape (n_samples,), weights for individual samples. default: None
    zero_division: Which value to substitute as a metric value when encountering zero division. Should be on of 0, 1,
        "warn". "warn" acts as 0, but the warning is raised.

Returns:
    'scores': dict. Summary of the scores for overall and per type
        Overall:
            'accuracy': accuracy,
            'precision': precision,
            'recall': recall,
            'f1': F1 score, also known as balanced F-score or F-measure,
        Per type:
            'precision': precision,
            'recall': recall,
            'f1': F1 score, also known as balanced F-score or F-measure
Examples:

    >>> predictions = [['O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
    >>> references = [['O', 'O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
    >>> seqeval = jury.load_metric("seqeval")
    >>> results = seqeval.compute(predictions=predictions, references=references)
    >>> print(list(results.keys()))
    ['MISC', 'PER', 'overall_precision', 'overall_recall', 'overall_f1', 'overall_accuracy']
    >>> print(results)
    {
      "seqeval": {
        "MISC": {
          "precision": 0.0,
          "recall": 0.0,
          "f1": 0.0,
          "number": 1
        },
        "PER": {
          "precision": 1.0,
          "recall": 1.0,
          "f1": 1.0,
          "number": 1
        },
        "overall_precision": 0.5,
        "overall_recall": 0.5,
        "overall_f1": 0.5,
        "overall_accuracy": 0.8
      }
    }
"""


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class SeqevalForLanguageGeneration(MetricForSequenceLabeling):
    def _download_and_prepare(self, dl_manager):
        global seqeval
        try:
            import seqeval.metrics
        except ModuleNotFoundError:
            raise ModuleNotFoundError(requirement_message(path="Seqeval", package_name="seqeval"))
        else:
            super(SeqevalForLanguageGeneration, self)._download_and_prepare(dl_manager)

    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            homepage="https://github.com/chakki-works/seqeval",
            inputs_description=_KWARGS_DESCRIPTION,
            features=self._default_features,
            codebase_urls=["https://github.com/chakki-works/seqeval"],
            reference_urls=["https://github.com/chakki-works/seqeval"],
        )

    def _compute_single_pred_single_ref(
        self,
        predictions: SequenceLabelingInstance,
        references: SequenceLabelingInstance,
        suffix: bool = False,
        scheme: Optional[str] = None,
        mode: Optional[str] = None,
        sample_weight: Optional[List[int]] = None,
        zero_division: Union[str, int] = "warn",
        json_normalized: bool = False,
    ):
        if scheme is not None:
            try:
                scheme_module = importlib.import_module("seqeval.scheme")
                scheme = getattr(scheme_module, scheme)
            except AttributeError:
                raise ValueError(f"Scheme should be one of [IOB1, IOB2, IOE1, IOE2, IOBES, BILOU], got {scheme}")
        report = seqeval.metrics.classification_report(
            y_true=references,
            y_pred=predictions,
            suffix=suffix,
            output_dict=True,
            scheme=scheme,
            mode=mode,
            sample_weight=sample_weight,
            zero_division=zero_division,
        )
        report.pop("macro avg")
        report.pop("weighted avg")
        overall_score = report.pop("micro avg")

        scores = {
            type_name: {
                "precision": float(score["precision"]),
                "recall": float(score["recall"]),
                "f1": float(score["f1-score"]),
                "number": int(score["support"]),
            }
            for type_name, score in report.items()
        }

        if json_normalized:
            scores = self.json_normalize(scores)

        scores["overall_precision"] = float(overall_score["precision"])
        scores["overall_recall"] = float(overall_score["recall"])
        scores["overall_f1"] = float(overall_score["f1-score"])
        scores["overall_accuracy"] = float(seqeval.metrics.accuracy_score(y_true=references, y_pred=predictions))

        return scores

    def _compute_single_pred_multi_ref(
        self,
        predictions: SequenceLabelingInstance,
        references: SequenceLabelingInstance,
        **kwargs,
    ):
        return self._compute_single_pred_single_ref(predictions=predictions, references=references, **kwargs)

    def _compute_multi_pred_multi_ref(
        self,
        predictions: SequenceLabelingInstance,
        references: SequenceLabelingInstance,
        **kwargs,
    ):
        return self._compute_single_pred_single_ref(predictions=predictions, references=references, **kwargs)

    @staticmethod
    def json_normalize(scores: Dict) -> Dict[str, float]:
        normalized = {}
        for key, value in scores.items():
            if isinstance(value, dict):
                for name, val in value.items():
                    normalized[f"{key}_{name}"] = val
            else:
                normalized[key] = value
        return normalized



from jury.metrics._core import MetricAlias
from jury.metrics.wer.wer_for_language_generation import WERForLanguageGeneration

__main_class__ = "WER"


class WER(MetricAlias):
    _SUBCLASS = WERForLanguageGeneration



from jury.metrics._core import TaskMapper
from jury.metrics.recall.recall_for_language_generation import RecallForLanguageGeneration
from jury.metrics.recall.recall_for_sequence_classification import RecallForSequenceClassification

__main_class__ = "Recall"


class Recall(TaskMapper):
    _TASKS = {
        "language-generation": RecallForLanguageGeneration,
        "sequence-classification": RecallForSequenceClassification,
    }



from jury.metrics._core import TaskMapper
from jury.metrics.f1.f1_for_language_generation import F1ForLanguageGeneration
from jury.metrics.f1.f1_for_sequence_classification import F1ForSequenceClassification

__main_class__ = "F1"


class F1(TaskMapper):
    _TASKS = {"language-generation": F1ForLanguageGeneration, "sequence-classification": F1ForSequenceClassification}



from jury.metrics._core import MetricAlias
from jury.metrics.chrf.chrf_for_language_generation import CHRFForLanguageGeneration

__main_class__ = "CHRF"


class CHRF(MetricAlias):
    _SUBCLASS = CHRFForLanguageGeneration



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Accuracy metric. The part of this file is adapted from HuggingFace's
datasets package implementation of Accuracy metric. See
https://github.com/huggingface/datasets/blob/master/metrics/accuracy/accuracy.py
"""

from collections import Counter
from typing import Callable

import datasets
import numpy as np

from jury.collator import Collator
from jury.metrics._core import MetricForLanguageGeneration
from jury.utils.nlp import normalize_text

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""

_DESCRIPTION = """
Accuracy is the proportion of correct predictions among the total number of cases processed. It can be computed with:
Accuracy = (TP + TN) / (TP + TN + FP + FN)
TP: True positive
TN: True negative
FP: False positive
FN: False negative

Accuracy for language generation computes token based accuracy.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: list of predictions to score. Each predictions
        should be a string with tokens separated by spaces.
    references: list of reference for each prediction. Each
        reference should be a string with tokens separated by spaces.
Returns:
    'score': Accuracy score.
Examples:

    >>> accuracy = jury.load_metric("accuracy")
    >>> predictions = [["the cat is on the mat", "There is cat playing on the mat"], ["Look! a wonderful day."]]
    >>> references = [
        ["the cat is playing on the mat.", "The cat plays on the mat."], 
        ["Today is a wonderful day", "The weather outside is wonderful."]
    ]
    >>> results = accuracy.compute(predictions=predictions, references=references)
    >>> print(results)
    {'accuracy': {'score': 0.7285714285714285}}
"""


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class AccuracyForLanguageGeneration(MetricForLanguageGeneration):
    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=self._default_features,
            reference_urls=["https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html"],
        )

    @staticmethod
    def _tokenize(predictions: Collator, references: Collator):
        predictions = [normalize_text(p).split() for p in predictions]
        references = [normalize_text(r).split() for r in references]
        return predictions, references

    def _compute_single_pred_single_ref(
        self, predictions: Collator, references: Collator, reduce_fn: Callable = None, **kwargs
    ):
        scores = []
        predictions, references = self._tokenize(predictions, references)
        for pred, ref in zip(predictions, references):
            score = 0
            pred_counts = Counter(pred)
            ref_counts = Counter(ref)
            for token, pred_count in pred_counts.items():
                if token in ref_counts:
                    score += min(pred_count, ref_counts[token])  # Intersection count
            scores.append(score / max(len(pred), len(ref)))
        avg_score = sum(scores) / len(scores)
        return {"score": avg_score}

    def _compute_single_pred_multi_ref(
        self, predictions: Collator, references: Collator, reduce_fn: Callable = None, **kwargs
    ):
        scores = []
        for pred, refs in zip(predictions, references):
            pred_score = [
                self._compute_single_pred_single_ref(Collator([pred], keep=True), Collator([ref], keep=True))
                for ref in refs
            ]
            reduced_score = self._reduce_scores(pred_score, reduce_fn=reduce_fn)
            scores.append(reduced_score)

        return self._reduce_scores(scores, reduce_fn=np.mean)

    def _compute_multi_pred_multi_ref(
        self, predictions: Collator, references: Collator, reduce_fn: Callable = None, **kwargs
    ):
        scores = []
        for preds, refs in zip(predictions, references):
            pred_scores = []
            for pred in preds:
                pred_score = self._compute_single_pred_multi_ref(
                    Collator([pred], keep=True), Collator([refs], keep=True), reduce_fn=reduce_fn
                )
                pred_scores.append(pred_score)
            reduced_score = self._reduce_scores(pred_scores, reduce_fn=reduce_fn)
            scores.append(reduced_score)

        return self._reduce_scores(scores, reduce_fn=np.mean)



import importlib.util
import os
import warnings
from pathlib import Path
from typing import Callable, Sequence, Union

import numpy as np
import requests

PACKAGE_CORE = Path(os.path.abspath(os.path.dirname(__file__)))
METRICS_ROOT = PACKAGE_CORE.parent
PACKAGE_SOURCE = METRICS_ROOT.parent
PROJECT_ROOT = PACKAGE_SOURCE.parent


class PackagePlaceholder:
    def __init__(self, version: str):
        self.__version__ = version


class TaskNotAvailable(KeyError):
    def __init__(self, path: str, task: str):
        message = f"Task '{task}' is not available for metric '{path}'."
        self.message = message
        super(TaskNotAvailable, self).__init__(message)


def requirement_message(path: str, package_name: str, package_source: str = None) -> str:
    package_source = package_source or package_name
    return (
        f"In order to use metric '{path}', '{package_name}' is required. "
        f"You can install the package by `pip install {package_source}`."
    )


def download(source: str, destination: str, overwrite: bool = False, warn: bool = False) -> None:
    if os.path.exists(destination) and not overwrite:
        if warn:
            warnings.warn(
                f"Path {destination} already exists, not overwriting. To overwrite, speficy 'overwrite' parameter."
            )
        return
    r = requests.get(source, allow_redirects=True)

    with open(destination, "wb") as out_file:
        out_file.write(r.content)


def import_module(module_name: str, filepath: str):
    spec = importlib.util.spec_from_file_location(module_name, filepath)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def get_token_lengths(sequences: Sequence[Sequence[str]], reduce_fn: Callable = None) -> Union[int, Sequence[int]]:
    token_lengths = [len(item) for item in sequences]
    if reduce_fn is not None:
        return int(reduce_fn(token_lengths))
    return token_lengths


def is_reduce_fn(fun: Callable) -> bool:
    result = np.array(fun([1, 2]))
    return result.size == 1


def list_metrics():
    _internal_metrics_path = METRICS_ROOT
    metric_modules = list(_internal_metrics_path.glob("[!_]*"))
    return [module_name.name.replace(".py", "") for module_name in metric_modules]



from jury.metrics._core.auto import AutoMetric, load_metric
from jury.metrics._core.auxiliary import MetricAlias, TaskMapper
from jury.metrics._core.base import (
    EvaluationInstance,
    LanguageGenerationInstance,
    Metric,
    MetricForCrossLingualEvaluation,
    MetricForLanguageGeneration,
    MetricForSequenceClassification,
    MetricForSequenceLabeling,
    MetricForTask,
    MetricOutput,
    SequenceClassificationInstance,
    SequenceLabelingInstance,
)
from jury.metrics._core.utils import PROJECT_ROOT, list_metrics



from typing import Any, Dict, Optional, Union

from jury.metrics._core.base import MetricForTask
from jury.metrics._core.utils import TaskNotAvailable
from jury.utils.common import camel_to_snake


class TaskMapper:
    """
    Base metric factory class which will be used as mapper for any metric class. This class is used by
    :py:class:`jury.AutoMetric` for loading specified metric. It maps the class to a specified metric class
    if multiple tasks are available for the metric.

    All metrics using TaskMapper must implement _TASKS attribute.

    Note:
        Use :py:class:`jury.metrics.TaskMapper` instead in case of metrics implementing a single task.
    """

    _TASKS: Dict[str, MetricForTask] = None

    def __init__(self, *args, **kwargs):
        raise EnvironmentError("This class is designed to be instantiated by using 'by_task()' method.")

    @classmethod
    def construct(
        cls, task: str, resulting_name: Optional[str] = None, compute_kwargs: Optional[Dict[str, Any]] = None, **kwargs
    ) -> MetricForTask:
        """
        Common interface for all metrics for specified MetricForTask to be constructed.

        Args:
            task: (``str``) Task name for the desired metric to obtain the subclass.
            resulting_name (Optional ``str``): Resulting name of the computed score returned. If None,
                `~._get_path()` is used.
            compute_kwargs (Optional ``Dict[str, Any]``): Arguments to be passed to `compute()` method of metric at
                computation.

        Raises: :py:class:`TaskNotAvailable`

        Returns: Metric for proper task if available.
        """
        subclass = cls._get_subclass(task=task)
        path = cls._get_path()
        resulting_name = resulting_name or path
        if subclass is None:
            raise TaskNotAvailable(path=path, task=task)
        return subclass._construct(resulting_name=resulting_name, compute_kwargs=compute_kwargs, **kwargs)

    @classmethod
    def _get_subclass(cls, task: str) -> Union[MetricForTask, None]:
        """
        All metric modules must implement this method as it is used to call metrics by default. Should raise
        proper exception (``TaskNotAvailable``) if the task is not supported by the metric.

        Args:
            task: (``str``) Task name for the desired metric.

        Returns: Metric for proper task if available, None otherwise.
        """
        return cls._TASKS.get(task, None)

    @classmethod
    def _get_path(cls):
        return camel_to_snake(cls.__name__)


class MetricAlias(TaskMapper):
    """
    Extension of TaskMapper which allows a single :py:class:`jury.metrics.MetricForTask` class to be aliased. If a
    metric has a single task, use this class instead of :py:class:`jury.metrics._core.TaskMapper`.

    All metrics using MetricAlias must implement _SUBCLASS attribute.
    """

    _SUBCLASS: MetricForTask = None

    @classmethod
    def construct(
        cls,
        task: str = None,
        resulting_name: Optional[str] = None,
        compute_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> MetricForTask:
        """
        Common interface for all metrics for specified MetricForTask to be constructed. Do not raise
        :py:class:`TaskNotAvailable` unlike :py:class:`TaskMapper` as it directly uses _SUBCLASS defined.

        Args:
            task: (Ignored ``str``) Ignored. Preserved to provide a common interface.
            resulting_name (Optional ``str``): Resulting name of the computed score returned. If None,
                `~._get_path()` is used.
            compute_kwargs (Optional ``Dict[str, Any]``): Arguments to be passed to `compute()` method of metric at
                computation.

        Returns: Metric for proper task if available.
        """
        subclass = cls._get_subclass()
        resulting_name = resulting_name or cls._get_path()
        return subclass._construct(resulting_name=resulting_name, compute_kwargs=compute_kwargs, **kwargs)

    @classmethod
    def _get_subclass(cls, *args, **kwargs) -> MetricForTask:
        return cls._SUBCLASS



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
BLEURT metric. The part of this file is adapted from BLEURT implementation
of datasets package. See
https://github.com/huggingface/datasets/blob/master/metrics/bleurt/bleurt.py
"""

import os
from typing import Callable

import datasets

from jury.metrics import LanguageGenerationInstance, MetricForLanguageGeneration
from jury.metrics._core.utils import PackagePlaceholder, requirement_message

# `import bleurt` placeholder
bleurt = PackagePlaceholder(version="1.2.2")

logger = datasets.logging.get_logger(__name__)


_CITATION = """\
@inproceedings{bleurt,
  title={BLEURT: Learning Robust Metrics for Text Generation},
  author={Thibault Sellam and Dipanjan Das and Ankur P. Parikh},
  booktitle={ACL},
  year={2020},
  url={https://arxiv.org/abs/2004.04696}
}
"""

_DESCRIPTION = """\
BLEURT a learnt evaluation metric for Natural Language Generation. It is built using multiple phases of transfer 
learning starting from a pretrained BERT model (Devlin et al. 2018) and then employing another pre-training 
phrase using synthetic data. Finally it is trained on WMT human annotations. You may run BLEURT 
out-of-the-box or fine-tune it for your specific application (the latter is expected to perform better).

See the project's README at https://github.com/google-research/bleurt#readme for more information.
"""

_KWARGS_DESCRIPTION = """
BLEURT score.

Args:
    `predictions` (list of str): prediction/candidate sentences
    `references` (list of str): reference sentences
    `config_name` BLEURT checkpoint. Will default to BLEURT-base if None.

Returns:
    'scores': List of scores.
Examples:

    >>> predictions = [["the cat is on the mat", "There is cat playing on the mat"], ["Look! a wonderful day."]]
    >>> references = [
        ["the cat is playing on the mat.", "The cat plays on the mat."], 
        ["Today is a wonderful day", "The weather outside is wonderful."]
    ]
    >>> bleurt = jury.load_metric("bleurt", config_name="bleurt-tiny-128")
    >>> results = bleurt.compute(predictions=predictions, references=references)
    >>> print(results)
    {'bleurt': {'score': 0.25963682122528553, 'scores': [0.47344309091567993, 0.04583055153489113], 'checkpoint': 'bleurt-tiny-128'}}
"""

_LICENSE = """Copyright 2021 Google.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

CHECKPOINT_URLS = {
    "bleurt-tiny-128": "https://storage.googleapis.com/bleurt-oss/bleurt-tiny-128.zip",
    "bleurt-tiny-512": "https://storage.googleapis.com/bleurt-oss/bleurt-tiny-512.zip",
    "bleurt-base-128": "https://storage.googleapis.com/bleurt-oss/bleurt-base-128.zip",
    "bleurt-base-512": "https://storage.googleapis.com/bleurt-oss/bleurt-base-512.zip",
    "bleurt-large-128": "https://storage.googleapis.com/bleurt-oss/bleurt-large-128.zip",
    "bleurt-large-512": "https://storage.googleapis.com/bleurt-oss/bleurt-large-512.zip",
    "BLEURT-20-D3": "https://storage.googleapis.com/bleurt-oss-21/BLEURT-20-D3.zip",
    "BLEURT-20-D6": "https://storage.googleapis.com/bleurt-oss-21/BLEURT-20-D6.zip",
    "BLEURT-20-D12": "https://storage.googleapis.com/bleurt-oss-21/BLEURT-20-D12.zip",
    "BLEURT-20": "https://storage.googleapis.com/bleurt-oss-21/BLEURT-20.zip",
}


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class BleurtForLanguageGeneration(MetricForLanguageGeneration):
    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            homepage="https://github.com/google-research/bleurt",
            inputs_description=_KWARGS_DESCRIPTION,
            features=self._default_features,
            codebase_urls=["https://github.com/google-research/bleurt"],
            reference_urls=["https://github.com/google-research/bleurt", "https://arxiv.org/abs/2004.04696"],
            license=_LICENSE,
        )

    def _download_and_prepare(self, dl_manager):
        global bleurt
        try:
            from bleurt import score
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                requirement_message(
                    path="bleurt",
                    package_name="bleurt",
                    package_source="git+https://github.com/devrimcavusoglu/bleurt.git",
                )
            )

        # check that config name specifies a valid BLEURT model
        if self.config_name == "default":
            logger.warning(
                "Using default BLEURT-Base checkpoint for sequence maximum length 128. "
                "You can use a bigger model for better results with e.g: "
                "Jury(metric={path: 'bleurt', 'checkpoint': 'bleurt-large-512'})."
            )
            self.config_name = "bleurt-base-128"

        if self.config_name.lower() in CHECKPOINT_URLS:
            checkpoint_name = self.config_name.lower()

        elif self.config_name.upper() in CHECKPOINT_URLS:
            checkpoint_name = self.config_name.upper()

        else:
            raise KeyError(
                f"{self.config_name} model not found. You should supply the name of a model checkpoint for bleurt in {CHECKPOINT_URLS.keys()}"
            )

        # download the model checkpoint specified by self.config_name and set up the scorer
        model_path = dl_manager.download_and_extract(CHECKPOINT_URLS[checkpoint_name])
        self.scorer = score.BleurtScorer(os.path.join(model_path, checkpoint_name))

    def _compute_single_pred_single_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        **kwargs,
    ):
        scores = self.scorer.score(references=references, candidates=predictions)
        return {"score": sum(scores) / len(scores), "scores": scores, "checkpoint": self.config_name}

    def _compute_single_pred_multi_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        **kwargs,
    ):
        scores = []
        for pred, refs in zip(predictions, references):
            pred = [pred] * len(refs)
            pred_scores = self.scorer.score(references=refs, candidates=pred)
            reduced_score = float(reduce_fn(pred_scores))
            scores.append(reduced_score)
        return {"score": sum(scores) / len(scores), "scores": scores, "checkpoint": self.config_name}

    def _compute_multi_pred_multi_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        **kwargs,
    ):
        scores = []
        for preds, refs in zip(predictions, references):
            pred_scores = []
            for pred in preds:
                pred = [pred] * len(refs)
                pred_score = self.scorer.score(references=refs, candidates=pred)
                pred_scores.append(reduce_fn(pred_score))
            reduced_score = float(reduce_fn(pred_scores))
            scores.append(reduced_score)

        return {"score": sum(scores) / len(scores), "scores": scores, "checkpoint": self.config_name}


if __name__ == "__main__":
    import json

    predictions = [["hello there", "general kenobi"]]
    references = [["hello there", "general kenobi"]]
    bleurt = BleurtForLanguageGeneration()
    results = bleurt.compute(predictions=predictions, references=references)
    print(json.dumps(results, indent=2))



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Accuracy metric. The part of this file is adapted from HuggingFace's
datasets package implementation of Accuracy metric. See
https://github.com/huggingface/datasets/blob/master/metrics/accuracy/accuracy.py
"""
from typing import Callable

import datasets
from sklearn.metrics import accuracy_score

from jury.metrics._core import MetricForSequenceClassification, SequenceClassificationInstance

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""

_DESCRIPTION = """
Accuracy is the proportion of correct predictions among the total number of cases processed. It can be computed with:
Accuracy = (TP + TN) / (TP + TN + FP + FN)
TP: True positive
TN: True negative
FP: False positive
FN: False negative
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Predicted labels, as returned by a model.
    references: Ground truth labels.
    normalize: If False, return the number of correctly classified samples.
        Otherwise, return the fraction of correctly classified samples.
    sample_weight: Sample weights.
Returns:
    accuracy: Accuracy score.
Examples:

    >>> accuracy = jury.load_metric("accuracy", task="sequence-classification")
    >>> predictions = [[0], [2], [1], [0], [0], [1]]
    >>> references = [[0], [1], [2], [0], [1], [2]]
    >>> results = accuracy.compute(predictions=predictions, references=references)
    >>> print(results)
    {'accuracy': {'score': 0.3333333333333333}}
"""


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class AccuracyForSequenceClassification(MetricForSequenceClassification):
    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=self._default_features,
            reference_urls=["https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html"],
        )

    def _compute_single_pred_single_ref(
        self,
        predictions: SequenceClassificationInstance,
        references: SequenceClassificationInstance,
        reduce_fn: Callable = None,
        normalize=True,
        sample_weight=None,
    ):
        accuracy = float(
            accuracy_score(y_pred=predictions, y_true=references, normalize=normalize, sample_weight=sample_weight)
        )
        return {"score": accuracy}

    def _compute_single_pred_multi_ref(
        self,
        predictions: SequenceClassificationInstance,
        references: SequenceClassificationInstance,
        reduce_fn: Callable = None,
        exact_match=True,
    ):
        n_samples = len(predictions)
        match_sum = 0
        for pred, ref in zip(predictions, references):
            if exact_match and [pred] == ref:
                match_sum += 1
            elif not exact_match and pred in ref:
                match_sum += 1
        return {"score": match_sum / n_samples}

    def _compute_multi_pred_multi_ref(
        self,
        predictions: SequenceClassificationInstance,
        references: SequenceClassificationInstance,
        reduce_fn: Callable = None,
        exact_match=True,
    ):
        n_samples = len(predictions)
        match_sum = 0
        for preds, ref in zip(predictions, references):
            if exact_match and preds == ref:
                match_sum += 1
            else:
                for pred in preds:
                    if pred in ref:
                        match_sum += 1
                        break
        return {"score": match_sum / n_samples}



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
BARTScore metric. The part of this file is adapted from metric implementations
of datasets package. See
https://github.com/huggingface/datasets/blob/master/metrics/
"""
from typing import Callable, Dict, List

import datasets
import numpy as np

from jury.metrics import LanguageGenerationInstance
from jury.metrics._core import MetricForLanguageGeneration

_CITATION = """
@misc{yuan2021bartscore,
      title={BARTScore: Evaluating Generated Text as Text Generation}, 
      author={Weizhe Yuan and Graham Neubig and Pengfei Liu},
      year={2021},
      eprint={2106.11520},
      archivePrefix={arXiv},
      primaryClass={cs.CL}
}
"""

_DESCRIPTION = """
BARTScore formulates evaluating generated text as a text generation task from pre-trained language models. It
operationalizes this idea using BART, an encoder-decoder based pre-trained model. BARTScore is conceptually
simple and empirically effective.

See the `README.md` file at [https://github.com/neulab/BARTScore](https://github.com/neulab/BARTScore) for more
information.
"""

_KWARGS_DESCRIPTION = """
BARTScore metric arguments.

Construction Args:
    model_checkpoint (str): BARTScore checkpoint. Will default to bartscore-large-cnn.
    model_weights (str): Optional BARTScore weights, overrides the checkpoint weights.
    device (str): On which the contextual embedding model will be allocated on.
        If this argument is None, the model lives on cuda:0 if cuda is available.
    
Computation Args:
    predictions (list of str): Prediction/candidate sentences.
    references (list of str or list of list of str): Reference sentences.
    batch_size (int): BARTScore processing batch size.
    segment_scores (bool): If True, then score for each instance are returned separately. Otherwise,
        average score is returned.

Returns:
    'score': BARTScore loss.

Examples:

    >>> bartscore = jury.load_metric("bartscore")
    >>> predictions = [["the cat is on the mat", "There is cat playing on the mat"], ["Look! a wonderful day."]]
    >>> references = [
        ["the cat is playing on the mat.", "The cat plays on the mat."], 
        ["Today is a wonderful day", "The weather outside is wonderful."]
    ]
    >>> results = bartscore.compute(predictions=predictions, references=references)
    >>> print(results)
    {
      "bartscore": {
        "score": -2.201135754585266,
        "model_checkpoint": "bartscore-large-cnn",
        "model_weights": null,
        "segment_scores": false
      }
    }
"""

_LICENSE = """Copyright 2022 NeuLab.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

CHECKPOINT_URLS = {
    "bartscore-large-cnn": {
        "model_checkpoint": "facebook/bart-large-cnn",
        "model_weights": {
            "parabank2": "https://drive.google.com/uc?export=download&id=1_7JfF7KOInb7ZrxKHIigTMR4ChVET01m&confirm=t"
        },
    }
}


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class BartscoreForLanguageGeneration(MetricForLanguageGeneration):
    def __init__(
        self,
        resulting_name: str = None,
        compute_kwargs: Dict = None,
        model_checkpoint: str = "bartscore-large-cnn",
        model_weights: str = None,
        max_length: int = 1024,
        device: str = None,
        **kwargs,
    ):
        self.model_checkpoint = model_checkpoint
        self.model_weights = model_weights
        self.max_length = max_length
        self.device = device
        self.model_path = None
        self.scorer = None
        super().__init__(resulting_name=resulting_name, compute_kwargs=compute_kwargs, **kwargs)

    def _download_and_prepare(self, dl_manager) -> None:
        """
        Downloads and import the computation of BARTScore score from the implementation
        of BARTScore computation from neulab/BARTScore. The code is sourced from a specific
        commit on the master branch, in order to keep things stable. See
        https://github.com/neulab/BARTScore/blob/47b8341854e1b8be965b65480ce236b0c2f7543b/bart_score.py
        """

        if self.model_checkpoint.lower() in CHECKPOINT_URLS:
            checkpoint_name = self.model_checkpoint.lower()
        else:
            raise KeyError(
                f"{self.model_checkpoint} checkpoint not found. You should supply the name of a model checkpoint for BARTScore in {CHECKPOINT_URLS.keys()}"
            )
        model_checkpoint = CHECKPOINT_URLS[checkpoint_name]["model_checkpoint"]

        if self.model_weights:
            if self.model_weights.lower() in CHECKPOINT_URLS[checkpoint_name]["model_weights"]:
                weights_name = self.model_weights.lower()
            else:
                raise KeyError(
                    f"Weights named '{self.model_weights}' not found. You should supply the name of a model weight for BARTScore in "
                    + str(list(CHECKPOINT_URLS[checkpoint_name]["model_weights"].keys()))
                )
            model_path = CHECKPOINT_URLS[checkpoint_name]["model_weights"][weights_name]
        else:
            model_path = None

        bartscore_source = (
            "https://raw.githubusercontent.com/neulab/BARTScore/47b8341854e1b8be965b65480ce236b0c2f7543b/bart_score.py"
        )
        self.external_module_path = dl_manager.download(bartscore_source)
        BARTScorer = self._get_external_resource("bart_score", attr="BARTScorer")

        self.scorer = BARTScorer(device=self.device, max_length=self.max_length, checkpoint=model_checkpoint)

        if model_path is not None:
            model_dest = dl_manager.download(model_path)
            self.scorer.load(path=model_dest)

    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            homepage="https://github.com/neulab/BARTScore",
            inputs_description=_KWARGS_DESCRIPTION,
            features=self._default_features,
            codebase_urls=["https://github.com/neulab/BARTScore"],
            reference_urls=[
                "https://github.com/neulab/BARTScore",
                "https://arxiv.org/abs/2106.11520",
            ],
            license=_LICENSE,
        )

    def _compute_single_pred_single_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        batch_size: int = 4,
        segment_scores: bool = False,
        **kwargs,
    ):
        score = self.scorer.score(predictions, references, batch_size=batch_size, **kwargs)
        if not segment_scores:
            score = float(np.mean(score))

        return {
            "score": score,
            "model_checkpoint": self.model_checkpoint,
            "model_weights": self.model_weights,
            "segment_scores": segment_scores,
        }

    def _compute_single_pred_multi_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        batch_size: int = 4,
        segment_scores: bool = False,
        **kwargs,
    ):
        """
        Ideally we would use the already-implemented multi-ref function in BARTScore,
        but unfortunately it only supports two reduction functions and doesn't support
        inconsistent number of reference per predictions. So instead we pre-combine our
        prediction/reference pairs and use their single prediction/reference function
        in order to utilize the batch_size argument.
        """
        flat_predictions: List[str] = []
        flat_references: List[str] = []
        ranges = []
        cursor = 0
        for pred, refs in zip(predictions, references):
            ref_count = len(refs)
            flat_predictions += [pred] * ref_count
            flat_references += refs
            new_cursor = cursor + ref_count
            ranges.append((cursor, new_cursor))
            cursor = new_cursor

        flat_scores = self.scorer.score(flat_predictions, flat_references, batch_size=batch_size, **kwargs)
        score = []
        for start, end in ranges:
            reduced_score = float(reduce_fn(flat_scores[start:end]))
            score.append(reduced_score)

        if not segment_scores:
            score = float(np.mean(score))

        return {
            "score": score,
            "model_checkpoint": self.model_checkpoint,
            "model_weights": self.model_weights,
            "segment_scores": segment_scores,
        }

    def _compute_multi_pred_multi_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        batch_size: int = 4,
        segment_scores: bool = False,
        **kwargs,
    ):
        """
        Like Single Pred/Multi Ref, we pre-combine all possible prediction/reference
        pairs into a list of single prediction/single reference pairs in order to utilize
        the batch_size argument.
        """
        flat_predictions: List[str] = []
        flat_references: List[str] = []
        ranges = []
        shapes = []
        cursor = 0
        for preds, refs in zip(predictions, references):
            pred_count = len(preds)
            ref_count = len(refs)
            for pred in preds:
                flat_predictions += [pred] * ref_count
                flat_references += refs
            new_cursor = cursor + (pred_count * ref_count)
            ranges.append((cursor, new_cursor))
            shapes.append((pred_count, ref_count))
            cursor = new_cursor

        flat_scores = self.scorer.score(flat_predictions, flat_references, batch_size=batch_size, **kwargs)
        score = []
        for pair_range, pair_shape in zip(ranges, shapes):
            pair_start, pair_end = pair_range
            pred_count, ref_count = pair_shape
            pair_scores = []
            for i in range(pred_count):
                pred_start = pair_start + i * ref_count
                pred_end = pred_start + ref_count
                pred_score = flat_scores[pred_start:pred_end]
                reduced_pred_score = float(reduce_fn(pred_score))
                pair_scores.append(reduced_pred_score)
            reduced_score = float(reduce_fn(pair_scores))
            score.append(reduced_score)

        if not segment_scores:
            score = float(np.mean(score))

        return {
            "score": score,
            "model_checkpoint": self.model_checkpoint,
            "model_weights": self.model_weights,
            "segment_scores": segment_scores,
        }



from jury.metrics._core import MetricAlias
from jury.metrics.cer.cer_for_language_generation import CERForLanguageGeneration

__main_class__ = "CER"


class CER(MetricAlias):
    _SUBCLASS = CERForLanguageGeneration



import pytest

from jury import Jury
from jury.metrics import AutoMetric
from tests.jury.conftest import get_expected_output
from tests.utils import assert_almost_equal_dict


@pytest.fixture(scope="module")
def jury_prism():
    metric = AutoMetric.load("prism")
    return Jury(metrics=metric)


@pytest.fixture
@get_expected_output(prefix="metrics")
def output_basic():
    return output_basic.output


@pytest.fixture
@get_expected_output(prefix="metrics")
def output_basic_normalized_and_segmented():
    return output_basic_normalized_and_segmented.output


@pytest.fixture
@get_expected_output(prefix="metrics")
def output_multiple_ref():
    return output_multiple_ref.output


@pytest.fixture
@get_expected_output(prefix="metrics")
def output_multiple_ref_normalized_and_segmented():
    return output_multiple_ref_normalized_and_segmented.output


@pytest.fixture
@get_expected_output(prefix="metrics")
def output_multiple_pred_multiple_ref():
    return output_multiple_pred_multiple_ref.output


@pytest.fixture
@get_expected_output(prefix="metrics")
def output_multiple_pred_multiple_ref_normalized_and_segmented():
    return output_multiple_pred_multiple_ref_normalized_and_segmented.output


def test_basic(predictions, references, jury_prism, output_basic):
    scores = jury_prism(predictions=predictions, references=references)
    assert_almost_equal_dict(actual=scores, desired=output_basic)


def test_basic_normalized_and_segmented(predictions, references, jury_prism, output_basic_normalized_and_segmented):
    scores = jury_prism(predictions=predictions, references=references, normalize=True, segment_scores=True)
    assert_almost_equal_dict(actual=scores, desired=output_basic_normalized_and_segmented)


def test_multiple_ref(predictions, multiple_references, jury_prism, output_multiple_ref):
    scores = jury_prism(predictions=predictions, references=multiple_references)
    assert_almost_equal_dict(actual=scores, desired=output_multiple_ref)


def test_multiple_ref_normalized_and_segmented(
    predictions, references, jury_prism, output_multiple_ref_normalized_and_segmented
):
    scores = jury_prism(predictions=predictions, references=references, normalize=True, segment_scores=True)
    assert_almost_equal_dict(actual=scores, desired=output_multiple_ref_normalized_and_segmented)


def test_multiple_pred_multiple_ref(
    multiple_predictions, multiple_references, jury_prism, output_multiple_pred_multiple_ref
):
    scores = jury_prism(predictions=multiple_predictions, references=multiple_references)
    assert_almost_equal_dict(actual=scores, desired=output_multiple_pred_multiple_ref)


def test_multiple_pred_multiple_ref_normalized_and_segmented(
    predictions,
    references,
    jury_prism,
    output_multiple_pred_multiple_ref_normalized_and_segmented,
):
    scores = jury_prism(predictions=predictions, references=references, normalize=True, segment_scores=True)
    assert_almost_equal_dict(actual=scores, desired=output_multiple_pred_multiple_ref_normalized_and_segmented)



import glob
import json
import os.path
from typing import Any, Dict, List, Optional, Tuple, Union

import fire
import pandas as pd

from jury import Jury
from jury import __version__ as jury_version
from jury.utils.common import get_common_keys
from jury.utils.io import json_load, json_save


def file_extension(path: str) -> str:
    return ".".join(path.split("/")[-1].split(".")[1:])


def from_file(
    predictions: str,
    references: str,
    reduce_fn: Optional[str] = None,
    config: Optional[str] = "",
    export: Optional[str] = None,
):
    args = json_load(config) or {}
    predictions = predictions or args.get("predictions")
    references = references or args.get("references")
    reduce_fn = reduce_fn or args.get("reduce_fn")
    metrics = args.get("metrics")
    scorer = Jury(metrics=metrics)

    if os.path.isfile(predictions) and os.path.isfile(references):
        scores = score_from_file(scorer=scorer, predictions=predictions, references=references, reduce_fn=reduce_fn)
    elif os.path.isdir(predictions) and os.path.isdir(references):
        paths = read_folders(predictions, references)
        scores = {}
        for pred_file, ref_file in paths:
            common_name = os.path.basename(pred_file)
            scores[common_name] = score_from_file(
                scorer=scorer, predictions=pred_file, references=ref_file, reduce_fn=reduce_fn
            )
    else:
        raise ValueError("predictions and references either both must be files or both must be folders.")

    if export:
        json_save(scores, export)

    print(json.dumps(scores, default=str, indent=4))


def read_file(filepath: str) -> Union[List[str], List[List[str]]]:
    if file_extension(filepath) == "csv":
        df = pd.read_csv(filepath, header=None)
        content = df.to_numpy().tolist()
    elif file_extension(filepath) == "tsv":
        df = pd.read_csv(filepath, header=None)
        content = df.to_numpy().tolist()
    else:
        with open(filepath, "r") as in_file:
            content = in_file.readlines()
    return content


def read_folders(predictions_path: str, references_path: str) -> List[Tuple[str, str]]:
    glob_predictions_path = os.path.join(predictions_path, "*")
    glob_references_path = os.path.join(references_path, "*")
    prediction_files = {os.path.basename(p): p for p in glob.glob(glob_predictions_path)}
    reference_files = {os.path.basename(p): p for p in glob.glob(glob_references_path)}

    common_files = get_common_keys(prediction_files, reference_files)

    files_to_read = []
    for common_file in common_files:
        common_pair = (prediction_files[common_file], reference_files[common_file])
        files_to_read.append(common_pair)

    return files_to_read


def score_from_file(scorer: Jury, predictions: str, references: str, reduce_fn: Optional[str] = None) -> Dict[str, Any]:
    predictions = read_file(predictions)
    references = read_file(references)
    return scorer(predictions=predictions, references=references, reduce_fn=reduce_fn)


def app() -> None:
    """Cli app."""
    fire.Fire({"version": jury_version, "eval": from_file})


if __name__ == "__main__":
    app()



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" CER metric. The part of this file is adapted from HuggingFace's
datasets package implementation of CER metric. See
https://github.com/huggingface/datasets/blob/master/metrics/cer/cer.py
"""
import warnings
from typing import Callable, List, Tuple, Union

import datasets

from jury.metrics import LanguageGenerationInstance, MetricForLanguageGeneration
from jury.metrics._core.utils import PackagePlaceholder, requirement_message

# `import jiwer` placeholder
jiwer = PackagePlaceholder(version="2.3.0")


_CITATION = """\
@inproceedings{inproceedings,
    author = {Morris, Andrew and Maier, Viktoria and Green, Phil},
    year = {2004},
    month = {01},
    pages = {},
    title = {From WER and RIL to MER and WIL: improved evaluation measures for connected speech recognition.}
}
"""

_DESCRIPTION = """\
Character error rate (CER) is a common metric of the performance of an automatic speech recognition system.
CER is similar to Word Error Rate (WER), but operates on character instead of word. Please refer to docs of WER for further information.
Character error rate can be computed as:
CER = (S + D + I) / N = (S + D + I) / (S + D + C)
where
S is the number of substitutions,
D is the number of deletions,
I is the number of insertions,
C is the number of correct characters,
N is the number of characters in the reference (N=S+D+C).
CER's output is not always a number between 0 and 1, in particular when there is a high number of insertions. This value is often associated to the percentage of characters that were incorrectly predicted. The lower the value, the better the
performance of the ASR system with a CER of 0 being a perfect score.
"""

_KWARGS_DESCRIPTION = """
Computes CER score of transcribed segments against references.
Args:
    references: list of references for each speech input.
    predictions: list of transcribtions to score.
    concatenate_texts: Whether or not to concatenate sentences before evaluation, set to True for more accurate result.
Returns:
    (float): the character error rate
Examples:
    >>> predictions = [["the cat is on the mat", "There is cat playing on the mat"], ["Look! a wonderful day."]]
    >>> references = [
        ["the cat is playing on the mat.", "The cat plays on the mat."], 
        ["Today is a wonderful day", "The weather outside is wonderful."]
    ]
    >>> cer = jury.load_metric("cer")
    >>> cer_score = cer.compute(predictions=predictions, references=references)
    >>> print(cer_score)
    {
      "cer": {
        "score": 0.7272727272727273,
        "overall": {
          "substitutions": 3.1666666666666665,
          "deletions": 5.5,
          "insertions": 2.3333333333333335,
          "hits": 19.166666666666668
        }
      }
    }
"""


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class CERForLanguageGeneration(MetricForLanguageGeneration):
    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=self._default_features,
            codebase_urls=["https://github.com/jitsi/jiwer/"],
            reference_urls=[
                "https://en.wikipedia.org/wiki/Word_error_rate",
                "https://sites.google.com/site/textdigitisation/qualitymeasures/computingerrorrates",
            ],
        )

    def _download_and_prepare(self, dl_manager):
        global jiwer
        global tr

        try:
            import jiwer
            import jiwer.transforms as tr
        except ModuleNotFoundError:
            raise ModuleNotFoundError(requirement_message(path="CER", package_name="jiwer"))
        else:
            super(CERForLanguageGeneration, self)._download_and_prepare(dl_manager)

    @staticmethod
    def _get_cer_transform():
        return tr.Compose(
            [
                tr.RemoveMultipleSpaces(),
                tr.Strip(),
                tr.ReduceToSingleSentence(""),
                tr.ReduceToListOfListOfChars(),
            ]
        )

    def _compute_cer_score(
        self, predictions: Union[str, List[str]], references: Union[str, List[str]]
    ) -> Tuple[float, int, int, int, int]:
        cer_transform = self._get_cer_transform()
        measures = jiwer.compute_measures(
            references,
            predictions,
            truth_transform=cer_transform,
            hypothesis_transform=cer_transform,
        )
        return (
            measures["wer"],
            int(measures["substitutions"]),
            int(measures["deletions"]),
            int(measures["insertions"]),
            int(measures["hits"]),
        )

    def _compute_single_pred_single_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        concatenate_texts: bool = False,
    ):
        if concatenate_texts:
            score, total_substitutions, total_deletions, total_insertions, total_hits = self._compute_cer_score(
                predictions, references
            )
        else:
            incorrect = 0
            total = 0
            total_substitutions = 0
            total_deletions = 0
            total_insertions = 0
            total_hits = 0
            for prediction, reference in zip(predictions, references):
                _, substitutions, deletions, insertions, hits = self._compute_cer_score(prediction, reference)
                total_substitutions += substitutions
                total_deletions += deletions
                total_insertions += insertions
                total_hits += hits
                incorrect += substitutions + deletions + insertions
                total += substitutions + deletions + hits
            score = incorrect / total

        return {
            "score": score,
            "overall": {
                "substitutions": total_substitutions,
                "deletions": total_deletions,
                "insertions": total_insertions,
                "hits": total_hits,
            },
        }

    def _compute_single_pred_multi_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        **kwargs
    ):
        if "concatenate_texts" in kwargs:
            warnings.warn("Option 'concatenate_texts' is only available in single-pred & single-ref setting.")

        total_substitutions = 0
        total_deletions = 0
        total_insertions = 0
        total_hits = 0
        total_refs = 0
        scores = []
        for pred, refs in zip(predictions, references):
            pred_scores = []
            for ref in refs:
                score, substitutions, deletions, insertions, hits = self._compute_cer_score(
                    predictions=pred, references=ref
                )
                pred_scores.append(score)
                total_substitutions += substitutions
                total_deletions += deletions
                total_insertions += insertions
                total_hits += hits
                total_refs += 1
            reduced_score = float(reduce_fn(pred_scores))
            scores.append(reduced_score)

        return {
            "score": float(reduce_fn(scores)),
            "overall": {
                "substitutions": total_substitutions / total_refs,
                "deletions": total_deletions / total_refs,
                "insertions": total_insertions / total_refs,
                "hits": total_hits / total_refs,
            },
        }

    def _compute_multi_pred_multi_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        **kwargs
    ):
        if "concatenate_texts" in kwargs:
            warnings.warn("Option 'concatenate_texts' is only available in single-pred & single-ref setting.")

        total_substitutions = 0
        total_deletions = 0
        total_insertions = 0
        total_hits = 0
        total_refs = 0
        scores = []
        for preds, refs in zip(predictions, references):
            pred_scores = []
            for pred in preds:
                for ref in refs:
                    score, substitutions, deletions, insertions, hits = self._compute_cer_score(
                        predictions=pred, references=ref
                    )
                    pred_scores.append(score)
                    total_substitutions += substitutions
                    total_deletions += deletions
                    total_insertions += insertions
                    total_hits += hits
                    total_refs += 1
            reduced_score = float(reduce_fn(pred_scores))
            scores.append(reduced_score)

        return {
            "score": float(reduce_fn(scores)),
            "overall": {
                "substitutions": total_substitutions / total_refs,
                "deletions": total_deletions / total_refs,
                "insertions": total_insertions / total_refs,
                "hits": total_hits / total_refs,
            },
        }



from jury.metrics._core import MetricAlias
from jury.metrics.sacrebleu.sacrebleu_for_language_generation import SacrebleuForLanguageGeneration

__main_class__ = "Sacrebleu"


class Sacrebleu(MetricAlias):
    _SUBCLASS = SacrebleuForLanguageGeneration



from jury.metrics._core import MetricAlias
from jury.metrics.ter.ter_for_language_generation import TERForLanguageGeneration

__main_class__ = "TER"


class TER(MetricAlias):
    _SUBCLASS = TERForLanguageGeneration



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Comet cross-lingual MT evaluation metric. The part of this file is adapted from
Comet implementation of datasets package. See
https://github.com/huggingface/datasets/blob/master/metrics/comet/comet.py
"""

from typing import Callable, Union

import datasets

from jury.metrics import LanguageGenerationInstance
from jury.metrics._core import MetricForCrossLingualEvaluation
from jury.metrics._core.utils import PackagePlaceholder, requirement_message

# `import comet` placeholder
comet = PackagePlaceholder(version="1.0.1")

_CITATION = """\
@inproceedings{rei-EtAl:2020:WMT,
   author    = {Rei, Ricardo  and  Stewart, Craig  and  Farinha, Ana C  and  Lavie, Alon},
   title     = {Unbabel's Participation in the WMT20 Metrics Shared Task},
   booktitle      = {Proceedings of the Fifth Conference on Machine Translation},
   month          = {November},
   year           = {2020},
   address        = {Online},
   publisher      = {Association for Computational Linguistics},
   pages     = {909--918},
}
@inproceedings{rei-etal-2020-comet,
   title = "{COMET}: A Neural Framework for {MT} Evaluation",
   author = "Rei, Ricardo  and
      Stewart, Craig  and
      Farinha, Ana C  and
      Lavie, Alon",
   booktitle = "Proceedings of the 2020 Conference on Empirical Methods in Natural Language Processing (EMNLP)",
   month = nov,
   year = "2020",
   address = "Online",
   publisher = "Association for Computational Linguistics",
   url = "https://www.aclweb.org/anthology/2020.emnlp-main.213",
   pages = "2685--2702",
}
"""

_DESCRIPTION = """\
Crosslingual Optimized Metric for Evaluation of Translation (COMET) is an open-source framework used to train 
Machine Translation metrics that achieve high levels of correlation with different types of human judgments 
(HTER, DA's or MQM). With the release of the framework the authors also released fully trained models that were used 
to compete in the WMT20 Metrics Shared Task achieving SOTA in that years competition. See the [README.md] file at 
https://unbabel.github.io/COMET/html/models.html for more information.
"""

_KWARGS_DESCRIPTION = """
COMET score.
Args:
`sources` (list of str): Source sentences
`predictions` (list of str): candidate translations
`references` (list of str): reference translations
`cuda` (bool): If set to True, runs COMET using GPU
`show_progress` (bool): Shows progress
`model`: COMET model to be used. Will default to `wmt-large-da-estimator-1719` if None.
Returns:
    `samples`: List of dictionaries with `src`, `mt`, `ref` and `score`.
    `scores`: List of scores.
Examples:
    >>> comet_metric = jury.load_metric('comet', config_name="wmt21-cometinho-da")
    >>> source = ["Die Katze spielt auf der Matte.", "Heute ist ein wunderbarer Tag."]
    >>> predictions = [["the cat is on the mat", "There is cat playing on the mat"], ["Look! a wonderful day."]]
    >>> references = [
        ["the cat is playing on the mat.", "The cat plays on the mat."], 
        ["Today is a wonderful day", "The weather outside is wonderful."]
    ]
    >>> results = comet_metric.compute(sources=source, predictions=hypothesis, references=reference)
    >>> print(results)
    {'comet': {'scores': [0.6338749527931213, 0.4925243854522705], 'samples': 0.5631996691226959}}
"""


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class CometForCrossLingualEvaluation(MetricForCrossLingualEvaluation):
    def _download_and_prepare(self, dl_manager):
        global comet
        try:
            import comet
        except ModuleNotFoundError:
            raise ModuleNotFoundError(requirement_message(path="comet", package_name="unbabel-comet"))
        else:
            super(CometForCrossLingualEvaluation, self)._download_and_prepare(dl_manager)

        if self.config_name == "default":
            checkpoint_path = comet.download_model("wmt20-comet-da")
        else:
            checkpoint_path = comet.download_model(self.config_name)
        self.scorer = comet.load_from_checkpoint(checkpoint_path)

    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            homepage="https://unbabel.github.io/COMET/html/index.html",
            inputs_description=_KWARGS_DESCRIPTION,
            features=self._default_features,
            codebase_urls=["https://github.com/Unbabel/COMET"],
            reference_urls=[
                "https://github.com/Unbabel/COMET",
                "https://www.aclweb.org/anthology/2020.emnlp-main.213/",
                "http://www.statmt.org/wmt20/pdf/2020.wmt-1.101.pdf6",
            ],
        )

    def _compute_single_pred_single_ref(
        self,
        sources: LanguageGenerationInstance,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        batch_size: int = 8,
        gpus: int = 1,
        mc_dropout: Union[int, bool] = False,
        progress_bar: bool = True,
        accelerator: str = "ddp",
        num_workers: int = None,
        length_batching: bool = True,
    ):
        data = {"src": sources, "mt": predictions, "ref": references}
        data = [dict(zip(data, t)) for t in zip(*data.values())]
        scores, samples = self.scorer.predict(
            data,
            batch_size=batch_size,
            gpus=gpus,
            mc_dropout=mc_dropout,
            progress_bar=progress_bar,
            accelerator=accelerator,
            num_workers=num_workers,
            length_batching=length_batching,
        )
        return {"scores": scores, "samples": samples}

    def _compute_single_pred_multi_ref(
        self,
        sources: LanguageGenerationInstance,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        batch_size: int = 8,
        gpus: int = 1,
        mc_dropout: Union[int, bool] = False,
        progress_bar: bool = True,
        accelerator: str = "ddp",
        num_workers: int = None,
        length_batching: bool = True,
    ):
        scores = []
        for src, pred, refs in zip(sources, predictions, references):
            data = {"src": [src] * len(refs), "mt": [pred] * len(refs), "ref": refs}
            data = [dict(zip(data, t)) for t in zip(*data.values())]
            pred_scores, _ = self.scorer.predict(
                data,
                batch_size=batch_size,
                gpus=gpus,
                mc_dropout=mc_dropout,
                progress_bar=progress_bar,
                accelerator=accelerator,
                num_workers=num_workers,
                length_batching=length_batching,
            )
            scores.append(float(reduce_fn(pred_scores)))

        return {"scores": scores, "samples": sum(scores) / len(scores)}

    def _compute_multi_pred_multi_ref(
        self,
        sources: LanguageGenerationInstance,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        batch_size: int = 8,
        gpus: int = 1,
        mc_dropout: Union[int, bool] = False,
        progress_bar: bool = True,
        accelerator: str = "ddp",
        num_workers: int = None,
        length_batching: bool = True,
    ):
        scores = []
        for src, preds, refs in zip(sources, predictions, references):
            all_pred_scores = []
            for pred in preds:
                data = {"src": [src] * len(refs), "mt": [pred] * len(refs), "ref": refs}
                data = [dict(zip(data, t)) for t in zip(*data.values())]
                pred_scores, _ = self.scorer.predict(
                    data,
                    batch_size=batch_size,
                    gpus=gpus,
                    mc_dropout=mc_dropout,
                    progress_bar=progress_bar,
                    accelerator=accelerator,
                    num_workers=num_workers,
                    length_batching=length_batching,
                )
                all_pred_scores.append(float(reduce_fn(pred_scores)))
            scores.append(float(reduce_fn(all_pred_scores)))

        return {"scores": scores, "samples": sum(scores) / len(scores)}



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
BLEU metric. The part of this file is adapted from HuggingFace's
datasets package implementation of BLEU metric. See
https://github.com/huggingface/datasets/blob/master/metrics/bleu/bleu.py
"""

import math
from typing import Callable, Dict, Tuple

import datasets

from jury.collator import Collator
from jury.metrics._core import MetricForLanguageGeneration
from jury.metrics._core.utils import get_token_lengths
from jury.tokenizer import DefaultTokenizer, TokenizerWrapper

_CITATION = """\
@INPROCEEDINGS{Papineni02bleu:a,
    author = {Kishore Papineni and Salim Roukos and Todd Ward and Wei-jing Zhu},
    title = {BLEU: a Method for Automatic Evaluation of Machine Translation},
    booktitle = {},
    year = {2002},
    pages = {311--318}
}
@inproceedings{lin-och-2004-orange,
    title = "{ORANGE}: a Method for Evaluating Automatic Evaluation Metrics for Machine Translation",
    author = "Lin, Chin-Yew  and
      Och, Franz Josef",
    booktitle = "{COLING} 2004: Proceedings of the 20th International Conference on Computational Linguistics",
    month = "aug 23{--}aug 27",
    year = "2004",
    address = "Geneva, Switzerland",
    publisher = "COLING",
    url = "https://www.aclweb.org/anthology/C04-1072",
    pages = "501--507",
}
"""

_DESCRIPTION = """\
BLEU (bilingual evaluation understudy) is an algorithm for evaluating the quality of text which has been 
machine-translated from one natural language to another. Quality is considered to be the correspondence between a 
machine's output and that of a human: "the closer a machine translation is to a professional human translation, the 
better it is" – this is the central idea behind BLEU. BLEU was one of the first metrics to claim a high correlation 
with human judgements of quality, and remains one of the most popular automated and inexpensive metrics.

Scores are calculated for individual translated segments—generally sentences—by comparing them with a set of good 
quality reference translations. Those scores are then averaged over the whole corpus to reach an estimate of the 
translation's overall quality. Intelligibility or grammatical correctness are not taken into account[citation needed].

BLEU's output is always a number between 0 and 1. This value indicates how similar the candidate text is to 
the reference texts, with values closer to 1 representing more similar texts. Few human translations will attain a 
score of 1, since this would indicate that the candidate is identical to one of the reference translations. For this 
reason, it is not necessary to attain a score of 1. Because there are more opportunities to match, adding additional
reference translations will increase the BLEU score.
"""

_KWARGS_DESCRIPTION = """
Computes BLEU score of translated segments against one or more references.
Args:
    predictions: list of translations to score.
        Each translation should be tokenized into a list of tokens.
    references: list of lists of references for each translation.
        Each reference should be tokenized into a list of tokens.
    max_order: Maximum n-gram order to use when computing BLEU score.
    smooth: Whether or not to apply Lin et al. 2004 smoothing.
Returns:
    'score': BLEU score,
    'precisions': geometric mean of n-gram precisions,
    'brevity_penalty': brevity penalty,
    'length_ratio': ratio of lengths,
    'translation_length': translation_length,
    'reference_length': reference_length
Examples:

    >>> bleu = jury.load_metric("bleu")
    >>> predictions = [["the cat is on the mat", "There is cat playing on the mat"], ["Look! a wonderful day."]]
    >>> references = [
        ["the cat is playing on the mat.", "The cat plays on the mat."], 
        ["Today is a wonderful day", "The weather outside is wonderful."]
    ]
    >>> results = bleu.compute(predictions=predictions, references=references)
    >>> print(results)
    {
      "bleu": {
        "score": 0.42370250917168295,
        "precisions": [
          0.8823529411764706,
          0.6428571428571429,
          0.45454545454545453,
          0.125
        ],
        "brevity_penalty": 1.0,
        "length_ratio": 1.0,
        "translation_length": 11,
        "reference_length": 11
      }
    }
"""


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class BleuForLanguageGeneration(MetricForLanguageGeneration):
    def __init__(self, resulting_name: str = None, compute_kwargs: Dict = None, **kwargs):
        self.should_change_resulting_name = resulting_name is None
        self.tokenizer = DefaultTokenizer()
        super().__init__(resulting_name=resulting_name, compute_kwargs=compute_kwargs, **kwargs)

    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("string", id="tokens"), id="sequence"),
                    "references": datasets.Sequence(datasets.Value("string", id="tokens"), id="sequence"),
                }
            ),
            codebase_urls=[
                "https://github.com/tensorflow/nmt/blob/0be864257a76c151eef20ea689755f08bc1faf4e/nmt/scripts/bleu.py"
            ],
            reference_urls=[
                "https://en.wikipedia.org/wiki/BLEU",
                "https://towardsdatascience.com/evaluating-text-output-in-nlp-bleu-at-your-own-risk-e8609665a213",
            ],
        )

    def _download_and_prepare(self, dl_manager) -> None:
        """
        Downloads and import the computation of bleu score from the implementation
        of BLEU computation from tensorflow/nmt. The code is sourced from a specific
        commit on the master branch, in order to keep things stable. See
        https://github.com/tensorflow/nmt/blob/0be864257a76c151eef20ea689755f08bc1faf4e/nmt/scripts/bleu.py
        """
        nmt_source = "https://raw.githubusercontent.com/tensorflow/nmt/0be864257a76c151eef20ea689755f08bc1faf4e/nmt/scripts/bleu.py"
        self.external_module_path = dl_manager.download(nmt_source)

    def _preprocess(self, predictions: Collator, references: Collator) -> Tuple[Collator, Collator]:
        tokenizer_wrapper = TokenizerWrapper(self.tokenizer)
        return tokenizer_wrapper.tokenize(predictions, references)

    def _compute_bleu_score(self, predictions: Collator, references: Collator, max_order=4, smooth=False):
        evaluation_fn = self._get_external_resource("nmt_bleu", attr="compute_bleu")
        score = evaluation_fn(
            reference_corpus=references, translation_corpus=predictions, max_order=max_order, smooth=smooth
        )
        (bleu, precisions, bp, ratio, translation_length, reference_length) = score
        return {
            "score": bleu,
            "precisions": precisions,
            "brevity_penalty": bp,
            "length_ratio": ratio,
            "translation_length": translation_length,
            "reference_length": reference_length,
        }

    def _compute_single_pred_single_ref(
        self,
        predictions: Collator,
        references: Collator,
        reduce_fn: Callable = None,
        max_order: int = 4,
        smooth: bool = False,
    ):
        # _compute_bleu_score expects a slightly different reference structure
        #   than the one we currently have for single prediction single reference
        references = [[r] for r in references]
        return self._compute_bleu_score(
            predictions=predictions, references=references, max_order=max_order, smooth=smooth
        )

    def _compute_single_pred_multi_ref(
        self,
        predictions: Collator,
        references: Collator,
        reduce_fn: Callable = None,
        max_order: int = 4,
        smooth: bool = False,
    ):
        # Bleu score inherently supports multiple references.
        # Bypassing _compute_single_pred_single_ref, as it does structure
        #   manipulation that isn't needed for here
        return self._compute_bleu_score(
            predictions=predictions, references=references, max_order=max_order, smooth=smooth
        )

    def _compute_multi_pred_multi_ref(
        self,
        predictions: Collator,
        references: Collator,
        reduce_fn: Callable = None,
        max_order: int = 4,
        smooth: bool = False,
    ):
        """
        For multiple predictions in the input, the resulting bleu score is computed using corpus level bleu
        as if it is the same for single prediction inputs (which is directly supported by evaluation function).
        The only difference however is that, since the score is calculated corpus level, `reduce_fn` is ignored.
        In other words, there is no aggregation for multiple predictions, and it is rather handled by the corpus
        level bleu score computation. After the pure blue score is calculated, parameters are changed accordingly,
        so the resulting predictions may not be a direct match of bleu score equivalent.

        Args:
            predictions: (Collator) Collator object of predictions.
            references: (Collator) Collator object of references.
            reduce_fn: (Callable) Ignored.
            max_order: (int) Maximum order of bleu precision counts (e.g n in BLEU-n), passed to evaluation function.
            smooth: (bool) Whether to apply smoothing or not, passed to evaluation function. False by default.
        """
        flattened_predictions = []
        matched_references = []
        adjusted_reference_length = adjusted_prediction_length = 0
        for preds, refs in zip(predictions, references):
            n_preds = len(preds)
            adjusted_reference_length += get_token_lengths(refs, reduce_fn=min)
            adjusted_prediction_length += get_token_lengths(preds, reduce_fn=max)
            flattened_predictions.extend([pred for pred in preds])
            matched_references.extend([refs] * n_preds)

        flattened_predictions = Collator(flattened_predictions, keep=True)
        matched_references = Collator(matched_references, keep=True)
        score = self._compute_single_pred_multi_ref(
            predictions=flattened_predictions,
            references=matched_references,
            reduce_fn=reduce_fn,
            max_order=max_order,
            smooth=smooth,
        )

        prediction_length, reference_length = score["translation_length"], score["reference_length"]
        ratio = prediction_length / reference_length
        adjusted_ratio = adjusted_prediction_length / adjusted_reference_length

        bleu_score = score["score"]
        if ratio > 1.0:
            adjusted_bp = 1.0
        else:
            bp = math.exp(1 - 1.0 / ratio)
            adjusted_bp = math.exp(1 - 1.0 / adjusted_ratio)
            bleu_score = bleu_score * (adjusted_bp / bp)

        score.update(
            {
                "score": bleu_score,
                "precisions": score["precisions"],
                "brevity_penalty": adjusted_bp,
                "length_ratio": adjusted_ratio,
                "translation_length": adjusted_prediction_length,
                "reference_length": adjusted_reference_length,
            }
        )

        return score

    def evaluate(
        self, predictions: Collator, references: Collator, reduce_fn: Callable = None, **kwargs
    ) -> Dict[str, float]:
        max_order = kwargs.get("max_order")
        if max_order is not None and self.should_change_resulting_name:
            self.resulting_name += f"_{max_order}"

        return super().evaluate(predictions=predictions, references=references, reduce_fn=reduce_fn, **kwargs)



import warnings
from concurrent.futures import ProcessPoolExecutor
from typing import Any, Callable, Dict, List, Mapping, Optional, Union

from jury.collator import Collator
from jury.definitions import DEFAULT_METRICS
from jury.metrics import EvaluationInstance, Metric, load_metric
from jury.utils.common import pop_item_from_dict, replace, set_env

MetricParam = Union[str, Metric, Dict[str, Any]]


class Jury:
    r"""
    Simple evaluation pipeline for text based metrics. By default it computes BLEU(n),
    METEOR, ROUGE-L metrics.

    Note:

        If ``predictions`` and ``references`` are given as list of instances, the order is recieved
        as prediction & reference pairs and evaluation is done by prioratizing the order.

    Examples:

    .. code-block:: python

        >>> # Example Evaluation
        >>> predictions = [
            ["the cat is on the mat", "there is playing cat on the mat."],
            ["Look! a wonderful day."]
        ]
        >>> references = [
            ["the cat is playing on the mat.", "The cat plays on the mat."],
            ["Today is a wonderful day", "The weather outside is wonderful."]
        ]
        >>> scorer = Jury()
        >>> results = scorer(predictions=predictions, references=references)
        >>> print(results)
        {'bleu_1': {"score": 0.6111111111111112, ...}, ..., 'meteor': {"score": 0.6470588235294118, ...}}
    """

    def __init__(
        self,
        metrics: Optional[Union[MetricParam, List[MetricParam]]] = None,
        run_concurrent=False,
    ):
        self.metrics = self._load_metrics(metrics)
        self._concurrent = run_concurrent

        # Sanity check
        self._validate_metrics()

    def __call__(
        self,
        *,
        predictions: EvaluationInstance = None,
        references: EvaluationInstance = None,
        reduce_fn: Optional[Union[str, Callable]] = None,
        **kwargs,
    ) -> Dict[str, float]:
        """Restricts positional arguments to prevent potential inconsistency between predictions and references."""
        if predictions is None or references is None:
            raise TypeError("Both predictions and references have to be passed.")
        if len(predictions) != len(references):
            raise ValueError("Lengths of predictions and references must be equal.")

        scores = dict()
        scores["total_items"] = len(references)
        scores["empty_items"] = self._remove_empty(predictions, references)

        if scores["total_items"] == scores["empty_items"]:
            warnings.warn(
                "At least one of the pairs are empty for all evaluation instances. No evaluation takes place."
            )
            return scores

        if self._concurrent:
            inputs_list = self._prepare_concurrent_inputs(predictions, references, reduce_fn, kwargs)
            set_env("TOKENIZERS_PARALLELISM", "true")
            with ProcessPoolExecutor() as executor:
                for score in executor.map(self._compute_single_score, inputs_list):
                    scores.update(score)
        else:
            for metric in self.metrics:
                inputs = (metric, predictions, references, reduce_fn, kwargs)
                score = self._compute_single_score(inputs)
                scores.update(score)

        return scores

    def _remove_empty(self, predictions: EvaluationInstance, references: EvaluationInstance):
        n_items = len(predictions)
        n_empty = 0
        for i in reversed(range(n_items)):
            if not isinstance(predictions[i], (float, int)) and not isinstance(references[i], (float, int)):
                if not predictions[i] or not references[i]:
                    predictions.pop(i)
                    references.pop(i)
                    n_empty += 1
        return n_empty

    def _load_single_metric(self, metric: Union[str, Metric]) -> List[Metric]:
        if isinstance(metric, str):
            metric = load_metric(metric)
        return [metric]

    def _load_multiple_metrics(self, metrics: Union[List[str], List[Dict[str, Any]], List[Metric]]) -> List[Metric]:
        for i, metric_param in enumerate(metrics):
            if isinstance(metric_param, str):
                path = metric_param
                metrics = replace(metrics, load_metric(path.lower()), i)
            elif isinstance(metric_param, dict):
                path = metric_param.pop("path")  # must be given
                task = pop_item_from_dict(metric_param, "task")
                resulting_name = pop_item_from_dict(metric_param, "resulting_name")
                compute_kwargs = pop_item_from_dict(metric_param, "compute_kwargs")
                kwargs = metric_param
                metrics = replace(
                    metrics,
                    load_metric(
                        path=path,
                        task=task,
                        resulting_name=resulting_name,
                        compute_kwargs=compute_kwargs,
                        **kwargs,
                    ),
                    i,
                )
            elif isinstance(metric_param, Metric):
                continue
        return metrics

    def _load_metrics(self, metrics: Union[MetricParam, List[MetricParam]]) -> List[Metric]:
        if metrics is None:
            metrics = DEFAULT_METRICS

        if isinstance(metrics, (str, Metric)):
            metrics = self._load_single_metric(metrics)
        elif isinstance(metrics, list):
            metrics = self._load_multiple_metrics(metrics)
        else:
            raise ValueError(f"Unknown input type {type(metrics)}")

        return metrics

    def _score_to_dict(self, score, name: str) -> Dict[str, float]:
        if isinstance(score, dict):
            return score

        return {name: score}

    def _compute_single_score(self, inputs) -> Mapping[str, float]:
        metric, predictions, references, reduce_fn, kwargs = inputs
        if isinstance(metric, Metric):
            predictions, references = Collator(predictions), Collator(references)
            score = metric.compute(predictions=predictions, references=references, reduce_fn=reduce_fn, **kwargs)
        else:
            metric.resulting_name = metric.name
            score = metric.compute(predictions=predictions, references=references, **kwargs)
            score = self._score_to_dict(score, name=metric.name)
        return score

    def _prepare_concurrent_inputs(self, predictions, references, reduce_fn, kwargs):
        inputs = []
        for metric in self.metrics:
            inputs.append((metric, predictions, references, reduce_fn, kwargs))
        return inputs

    def _validate_metrics(self):
        metrics = self.metrics
        if all([isinstance(metric, Metric) for metric in metrics]):
            task = metrics[0].task
            if not all([metric.task == task for metric in metrics]):
                raise ValueError(
                    "Given metrics are not suitable to be used together, metrics must be of same the task."
                )
        return True

    def add_metric(self, path: str, resulting_name: str = None, compute_kwargs: Dict = None) -> None:
        metric = load_metric(path, resulting_name=resulting_name, compute_kwargs=compute_kwargs)
        if metric not in self.metrics:
            self.metrics.append(metric)
            self._validate_metrics()

    def remove_metric(self, resulting_name: str, error: bool = True) -> None:
        for i, metric in enumerate(self.metrics):
            if metric.resulting_name == resulting_name:
                self.metrics.pop(i)
                return
        if error:
            # raise an error if resulting_metric is not found
            raise ValueError(f"Metric with resulting name {resulting_name} does not exists.")

    def evaluate(
        self,
        *,
        predictions: Union[List[str], List[List[str]]] = None,
        references: Union[List[str], List[List[str]]] = None,
        reduce_fn: Optional[Union[str, Callable]] = None,
    ) -> Dict[str, float]:
        """Returns __call__() method. For backward compatibility."""
        return self.__call__(predictions=predictions, references=references, reduce_fn=reduce_fn)



DEFAULT_METRICS = [
    {"path": "bleu", "compute_kwargs": {"max_order": 1}},
    {"path": "bleu", "compute_kwargs": {"max_order": 2}},
    {"path": "bleu", "compute_kwargs": {"max_order": 3}},
    {"path": "bleu", "compute_kwargs": {"max_order": 4}},
    {"path": "meteor"},
    {"path": "rouge"},
]



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import os
import warnings
from typing import Any, Dict, NamedTuple, Optional

import datasets

from jury.metrics._core.base import Metric
from jury.metrics._core.utils import import_module, list_metrics


def load_metric(
    path: str,
    resulting_name: Optional[str] = None,
    task: Optional[str] = None,
    compute_kwargs: Optional[Dict[str, Any]] = None,
    use_jury_only: bool = False,
    **kwargs,
) -> Metric:
    """Load a :py:class:`jury.metrics.Metric`. Alias for :py:class:`jury.metrics.AutoMetric.load()`.

    Args:

        path (``str``):
            path to the metric processing script with the metric builder. Can be either:
                - a local absolute or relative path to processing script or the directory containing the script,
                    e.g. ``'./metrics/rogue/rouge.py'``
                - a metric identifier on the HuggingFace datasets repo (list all available metrics with
                    ``jury.list_metrics()``) e.g. ``'rouge'`` or ``'bleu'``
        resulting_name (Optional ``str``): Resulting name of the computed score returned.
        task (Optional ``str``): Task name for the metric. "language-generation" by default.
        compute_kwargs (Optional ``Dict[str, Any]``): Arguments to be passed to `compute()` method of metric at
            computation.
        use_jury_only (``bool``): Whether to use jury metrics only or not. False by default.
        kwargs (Optional): Additional keyword arguments to be passed to :py:func:`datasets.load_metric`.

    Returns:
        `datasets.Metric`
    """
    return AutoMetric.load(
        path=path,
        resulting_name=resulting_name,
        task=task,
        use_jury_only=use_jury_only,
        compute_kwargs=compute_kwargs,
        **kwargs,
    )


class AutoMetric:
    """
    Instantiates the proper metric class from given parameters.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        raise EnvironmentError("This class is designed to be instantiated by using 'from_params()' method.")

    @classmethod
    def load(
        cls,
        path: str,
        task: Optional[str] = None,
        resulting_name: Optional[str] = None,
        compute_kwargs: Optional[Dict[str, Any]] = None,
        use_jury_only: bool = False,
        **kwargs,
    ) -> Metric:
        resolved_path = cls.resolve_metric_path(path)
        if task is None:
            task = "language-generation"

        # load the module, will catch ModuleNotFoundError if module cannot be loaded
        try:
            module_path = resolved_path.path
            if resolved_path.resolution == "external-module":
                module_name = os.path.basename(module_path)
                module = import_module(module_name=module_name, filepath=module_path)
            else:
                module = importlib.import_module(module_path)
        except ModuleNotFoundError:
            # Metric not in Jury
            if use_jury_only:
                raise ValueError(
                    f"Metric {resolved_path.path} is not available on jury, set use_jury_only=False to use"
                    f"additional metrics (e.g datasets metrics)."
                )
            warnings.warn(
                f"Metric {resolved_path.path} is not available on jury, falling back to datasets metric. "
                f"You may not fully utilize this metric for different input types, e.g multiple predictions "
                f"or multiple references."
            )
            metric = datasets.load_metric(resolved_path.path, **kwargs)
        else:
            # get the class, will raise AttributeError if class cannot be found
            factory_class = module.__main_class__
            klass = getattr(module, factory_class)
            metric = klass.construct(task=task, resulting_name=resulting_name, compute_kwargs=compute_kwargs, **kwargs)
        return metric

    @staticmethod
    def resolve_metric_path(path: str):
        class ResolvedName(NamedTuple):
            path: str
            resolution: str

        if path in list_metrics():
            path = path.lower()
            module_name = f"jury.metrics.{path}.{path}"
            return ResolvedName(path=module_name, resolution="internal-module")
        elif os.path.exists(path):
            # Get absolute path
            path = os.path.abspath(path)
            if not os.path.isdir(path):
                raise ValueError(
                    "Given 'path' must be a enclosing directory of metric. The names of the directory "
                    "(custom_metric/) and metric script (custom_metric.py) must be the same."
                )
            parent_dir, module_name = os.path.split(path)
            path = os.path.join(parent_dir, module_name, module_name + ".py")
            if not os.path.exists(path):
                raise FileNotFoundError(f"File {path} does not exists.")
            return ResolvedName(path=path, resolution="external-module")
        else:
            return ResolvedName(path=path, resolution="datasets")



import pytest

from jury import Jury
from jury.metrics import AutoMetric
from tests.jury.conftest import get_expected_output
from tests.utils import assert_almost_equal_dict


@pytest.fixture(scope="module")
def jury_comet():
    metric = AutoMetric.load(
        "comet",
        config_name="wmt21-cometinho-da",
        compute_kwargs={"gpus": 0, "num_workers": 0, "progress_bar": False, "batch_size": 2},
    )
    return Jury(metrics=metric)


@pytest.fixture(scope="module")
def comet_sources():
    return ["Die Katze spielt auf der Matte.", "Heute ist ein wunderbarer Tag."]


@pytest.fixture
@get_expected_output(prefix="metrics")
def output_basic():
    return output_basic.output


@pytest.fixture
@get_expected_output(prefix="metrics")
def output_multiple_ref():
    return output_multiple_ref.output


@pytest.fixture
@get_expected_output(prefix="metrics")
def output_multiple_pred_multiple_ref():
    return output_multiple_pred_multiple_ref.output


def test_basic(comet_sources, predictions, references, jury_comet, output_basic):
    scores = jury_comet(sources=comet_sources, predictions=predictions, references=references)
    assert_almost_equal_dict(actual=scores, desired=output_basic)


def test_multiple_ref(comet_sources, predictions, multiple_references, jury_comet, output_multiple_ref):
    scores = jury_comet(sources=comet_sources, predictions=predictions, references=multiple_references)
    assert_almost_equal_dict(actual=scores, desired=output_multiple_ref)


def test_multiple_pred_multiple_ref(
    comet_sources, multiple_predictions, multiple_references, jury_comet, output_multiple_pred_multiple_ref
):
    scores = jury_comet(sources=comet_sources, predictions=multiple_predictions, references=multiple_references)
    assert_almost_equal_dict(actual=scores, desired=output_multiple_pred_multiple_ref)



from pathlib import Path

import pytest

from tests.jury import TEST_DATA_DIR
from tests.jury.conftest import get_expected_output
from tests.utils import assert_almost_equal_dict, shell_capture


@pytest.fixture
@get_expected_output(prefix=None)
def output_cli_from_file():
    return output_cli_from_file.output


@pytest.fixture
@get_expected_output(prefix=None)
def output_cli_from_folder():
    return output_cli_from_folder.output


def test_cli_from_file(output_cli_from_file):
    predictions_file = str(Path(TEST_DATA_DIR) / "cli" / "from_file" / "predictions.txt")
    references_file = str(Path(TEST_DATA_DIR) / "cli" / "from_file" / "references.txt")

    cmd = f"jury eval --predictions {predictions_file} --references {references_file}"
    out = shell_capture(cmd)
    assert_almost_equal_dict(actual=out, desired=output_cli_from_file)


def test_cli_from_folder(output_cli_from_folder):
    predictions_folder = str(Path(TEST_DATA_DIR) / "cli" / "from_folder" / "predictions")
    references_folder = str(Path(TEST_DATA_DIR) / "cli" / "from_folder" / "references")

    cmd = f"jury eval --predictions {predictions_folder} --references {references_folder}"
    out = shell_capture(cmd)
    assert_almost_equal_dict(actual=out, desired=output_cli_from_folder)



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Prism metric. The part of this file is adapted from metric implementations
of datasets package. See
https://github.com/huggingface/datasets/blob/master/metrics/
"""
import os
from typing import Callable, Dict, List, Union

import datasets
import validators

from jury.metrics import LanguageGenerationInstance
from jury.metrics._core import MetricForLanguageGeneration

_CITATION = """
@inproceedings{thompson-post-2020-automatic,
    title={Automatic Machine Translation Evaluation in Many Languages via Zero-Shot Paraphrasing},
    author={Brian Thompson and Matt Post},
    year={2020},
    booktitle = "Proceedings of the 2020 Conference on Empirical Methods in Natural Language Processing (EMNLP)",
    month = nov,
    address = "Online",
    publisher = "Association for Computational Linguistics",
}
"""

_DESCRIPTION = """
Prism is an automatic MT metric which uses a sequence-to-sequence paraphraser to score MT system outputs 
conditioned on their respective human references. Prism uses a multilingual NMT model as a zero-shot paraphraser, 
which negates the need for synthetic paraphrase data and results in a single model which works in many languages.

See the `README.md` file at [https://github.com/thompsonb/prism](https://github.com/thompsonb/prism) for more
information.
"""

_KWARGS_DESCRIPTION = """
Prism metric arguments.

Construction Args:
    model_path_or_url (str): Path to the model directory or a URL of model file (tar).
    lang (str): Language of the sentences; required (e.g. 'en').
    temperature (float): Softmax temperature, where values >1.0 produce more uniform samples 
        and values <1.0 produce sharper samples.
    
Computation Args:
    predictions (list of str): Prediction/candidate sentences.
    references (list of str or list of list of str): Reference sentences.
    segment_scores (bool): If True, then score for each instance are returned separately. Otherwise,
        average score is returned.
    normalized (bool): If True, resulting score/scores are normalized with exponentiation by log base,
        which bounds the score range within [0,1] (still higher is better). 

Returns:
    'score': Prism score.

Examples:

    >>> prism = jury.load_metric("prism")
    >>> predictions = [["the cat is on the mat", "There is cat playing on the mat"], ["Look! a wonderful day."]]
    >>> references = [
        ["the cat is playing on the mat.", "The cat plays on the mat."], 
        ["Today is a wonderful day", "The weather outside is wonderful."]
    ]
    >>> results = prism.compute(predictions=predictions, references=references)
    >>> print(results)
    {
      "prism": {
        "score": -2.3071975708007812,
        "identifier": {
          "version": "0.1",
          "model": "m39v1",
          "seg_scores": "avg_log_prob",
          "sys_scores": "avg_log_prob",
          "log_base": 2,
          "temperature": 1.0
        },
        "model_path_or_url": "default",
        "lang": "en",
        "segment_scores": false,
        "normalized": false
      }
    }
"""

_LICENSE = """MIT License

Copyright (c) Brian Thompson

Portions of this software are copied from fairseq (https://github.com/pytorch/fairseq),
which is released under the MIT License and Copyright (c) Facebook, Inc. and its affiliates.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

CHECKPOINT_URLS = {
    "default": {
        "url": "https://huggingface.co/Devrim/prism-default/resolve/main/m39v1.tar",
        "model_dir": "m39v1",
    }
}


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class PrismForLanguageGeneration(MetricForLanguageGeneration):
    def __init__(
        self,
        resulting_name: str = None,
        compute_kwargs: Dict = None,
        model_path_or_url: str = "default",
        lang: str = "en",
        temperature: float = 1.0,
        **kwargs,
    ):
        self.model_path_or_url = model_path_or_url
        self.lang = lang
        self.temperature = temperature
        self.model_dir = None
        self.scorer = None
        super().__init__(resulting_name=resulting_name, compute_kwargs=compute_kwargs, **kwargs)

    @property
    def model_identifier(self):
        if self.scorer is not None:
            return self.scorer.identifier()

    def _download_model(self, dl_manager):
        if (
            self.model_path_or_url not in CHECKPOINT_URLS
            and not os.path.isdir(self.model_path_or_url)
            and not validators.url(self.model_path_or_url)
        ):
            raise ValueError("Provided 'model_path_or_url' neither points to an existing directory nor a valid URL.")
        elif os.path.isdir(self.model_path_or_url):
            self.model_dir = self.model_path_or_url
        else:
            if self.model_path_or_url in CHECKPOINT_URLS:
                model_source = CHECKPOINT_URLS[self.model_path_or_url]["url"]
                model_dir = CHECKPOINT_URLS[self.model_path_or_url]["model_dir"]
            else:
                model_source = self.model_path_or_url
                model_dir = os.path.basename(self.model_path_or_url).replace(".tar", "")
            extraction_dir = dl_manager.download_and_extract(model_source)
            self.model_dir = os.path.join(extraction_dir, model_dir)

    def _download_and_prepare(self, dl_manager) -> None:
        """
        Downloads and import the computation of Prism score from the implementation
        of Prism computation from thompsonb/prism. The code is sourced from a specific
        commit on the master branch, in order to keep things stable. See
        https://github.com/thompsonb/prism/blob/42e45a46d1c7924e98bceeed2ea81b31efcb6f9d/prism.py
        """
        self._download_model(dl_manager)
        prism_source = (
            "https://raw.githubusercontent.com/thompsonb/prism/42e45a46d1c7924e98bceeed2ea81b31efcb6f9d/prism.py"
        )
        self.external_module_path = dl_manager.download(
            prism_source,
        )

    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            homepage="https://github.com/thompsonb/prism",
            inputs_description=_KWARGS_DESCRIPTION,
            features=self._default_features,
            codebase_urls=["https://github.com/thompsonb/prism"],
            reference_urls=[
                "https://github.com/thompsonb/prism",
                "https://www.aclweb.org/anthology/2020.emnlp-main.8/",
            ],
            license=_LICENSE,
        )

    def _load_scorer(self):
        if self.scorer is None:
            Prism = self._get_external_resource("prism", attr="Prism")
            self.scorer = Prism(model_dir=self.model_dir, lang=self.lang, temperature=self.temperature)

    def _compute_prism_score(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        segment_scores: bool,
        **kwargs,
    ) -> Union[float, List[float]]:
        self._load_scorer()
        score = self.scorer.score(ref=references, cand=predictions, segment_scores=segment_scores, **kwargs)
        if segment_scores:
            return score.tolist()
        return float(score)

    @staticmethod
    def _normalize_score(score: Union[float, List[float]], exponent: float):
        if isinstance(score, float):
            return exponent ** score
        return [exponent ** s for s in score]

    def _compute_single_pred_single_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn=None,
        segment_scores: bool = False,
        normalize: bool = False,
        **kwargs,
    ):
        score = self._compute_prism_score(predictions, references, segment_scores=segment_scores, **kwargs)
        if normalize:
            score = self._normalize_score(score, self.model_identifier["log_base"])

        return {
            "score": score,
            "identifier": self.model_identifier,
            "model_path_or_url": self.model_path_or_url,
            "lang": self.lang,
            "segment_scores": segment_scores,
            "normalized": normalize,
        }

    def _compute_single_pred_multi_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        segment_scores: bool = False,
        normalize: bool = False,
        **kwargs,
    ):
        self._load_scorer()
        scores = []
        for pred, refs in zip(predictions, references):
            pred = [pred] * len(refs)
            prism_score = self._compute_prism_score(predictions=pred, references=refs, segment_scores=True, **kwargs)
            reduced_score = float(reduce_fn(prism_score))
            scores.append(reduced_score)

        if not segment_scores:
            scores = sum(scores) / len(scores)

        if normalize:
            scores = self._normalize_score(scores, self.model_identifier["log_base"])

        return {
            "score": scores,
            "identifier": self.model_identifier,
            "model_path_or_url": self.model_path_or_url,
            "lang": self.lang,
            "segment_scores": segment_scores,
            "normalized": normalize,
        }

    def _compute_multi_pred_multi_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        segment_scores: bool = False,
        normalize: bool = False,
        **kwargs,
    ):
        self._load_scorer()
        scores = []
        for preds, refs in zip(predictions, references):
            pred_scores = []
            for pred in preds:
                pred = [pred] * len(refs)
                prism_score = self._compute_prism_score(
                    predictions=pred, references=refs, segment_scores=True, **kwargs
                )
                reduced_pred_score = float(reduce_fn(prism_score))
                pred_scores.append(reduced_pred_score)
            reduced_score = float(reduce_fn(pred_scores))
            scores.append(reduced_score)

        if not segment_scores:
            scores = sum(scores) / len(scores)

        if normalize:
            scores = self._normalize_score(scores, self.model_identifier["log_base"])

        return {
            "score": scores,
            "identifier": self.model_identifier,
            "model_path_or_url": self.model_path_or_url,
            "lang": self.lang,
            "segment_scores": segment_scores,
            "normalized": normalize,
        }



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Word Error Ratio (WER) metric. The part of this file is adapted from HuggingFace's
datasets package implementation of CER metric. See
https://github.com/huggingface/datasets/blob/master/metrics/wer/wer.py
"""
import warnings
from typing import Callable, List, Tuple, Union

import datasets

from jury.metrics import LanguageGenerationInstance, MetricForLanguageGeneration
from jury.metrics._core.utils import PackagePlaceholder, requirement_message

# `import jiwer` placeholder
jiwer = PackagePlaceholder(version="2.3.0")


_CITATION = """\
@inproceedings{inproceedings,
    author = {Morris, Andrew and Maier, Viktoria and Green, Phil},
    year = {2004},
    month = {01},
    pages = {},
    title = {From WER and RIL to MER and WIL: improved evaluation measures for connected speech recognition.}
}
"""

_DESCRIPTION = """\
Word error rate (WER) is a common metric of the performance of an automatic speech recognition system.
The general difficulty of measuring performance lies in the fact that the recognized word sequence can have a different length from the reference word sequence (supposedly the correct one). The WER is derived from the Levenshtein distance, working at the word level instead of the phoneme level. The WER is a valuable tool for comparing different systems as well as for evaluating improvements within one system. This kind of measurement, however, provides no details on the nature of translation errors and further work is therefore required to identify the main source(s) of error and to focus any research effort.
This problem is solved by first aligning the recognized word sequence with the reference (spoken) word sequence using dynamic string alignment. Examination of this issue is seen through a theory called the power law that states the correlation between perplexity and word error rate.
Word error rate can then be computed as:
WER = (S + D + I) / N = (S + D + I) / (S + D + C)
where
S is the number of substitutions,
D is the number of deletions,
I is the number of insertions,
C is the number of correct words,
N is the number of words in the reference (N=S+D+C).
This value indicates the average number of errors per reference word. The lower the value, the better the
performance of the ASR system with a WER of 0 being a perfect score.
"""

_KWARGS_DESCRIPTION = """
Compute WER score of transcribed segments against references.
Args:
    references: List of references for each speech input.
    predictions: List of transcriptions to score.
    concatenate_texts (bool, default=False): Whether to concatenate all input texts or compute WER iteratively.
Returns:
    (float): the word error rate
Examples:
    >>> predictions = [["the cat is on the mat", "There is cat playing on the mat"], ["Look! a wonderful day."]]
    >>> references = [
        ["the cat is playing on the mat.", "The cat plays on the mat."], 
        ["Today is a wonderful day", "The weather outside is wonderful."]
    ]
    >>> wer = jury.load_metric("wer")
    >>> wer_score = wer.compute(predictions=predictions, references=references)
    >>> print(wer_score)
    {
      "wer": {
        "score": 1.0,
        "overall": {
          "substitutions": 2.8333333333333335,
          "deletions": 0.5,
          "insertions": 0.16666666666666666,
          "hits": 2.6666666666666665
        }
      }
    }
"""


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class WERForLanguageGeneration(MetricForLanguageGeneration):
    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=self._default_features,
            codebase_urls=["https://github.com/jitsi/jiwer/"],
            reference_urls=[
                "https://en.wikipedia.org/wiki/Word_error_rate",
                "https://sites.google.com/site/textdigitisation/qualitymeasures/computingerrorrates",
            ],
        )

    def _download_and_prepare(self, dl_manager):
        global jiwer

        try:
            import jiwer
        except ModuleNotFoundError:
            raise ModuleNotFoundError(requirement_message(path="WER", package_name="jiwer"))
        else:
            super(WERForLanguageGeneration, self)._download_and_prepare(dl_manager)

    def _compute_wer_score(
        self, predictions: Union[str, List[str]], references: Union[str, List[str]]
    ) -> Tuple[float, int, int, int, int]:
        measures = jiwer.compute_measures(references, predictions)
        return (
            measures["wer"],
            int(measures["substitutions"]),
            int(measures["deletions"]),
            int(measures["insertions"]),
            int(measures["hits"]),
        )

    def _compute_single_pred_single_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        concatenate_texts: bool = False,
    ):
        if concatenate_texts:
            score, total_substitutions, total_deletions, total_insertions, total_hits = self._compute_wer_score(
                predictions, references
            )
        else:
            incorrect = 0
            total = 0
            total_substitutions = 0
            total_deletions = 0
            total_insertions = 0
            total_hits = 0
            for prediction, reference in zip(predictions, references):
                _, substitutions, deletions, insertions, hits = self._compute_wer_score(prediction, reference)
                total_substitutions += substitutions
                total_deletions += deletions
                total_insertions += insertions
                total_hits += hits
                incorrect += substitutions + deletions + insertions
                total += substitutions + deletions + hits
            score = incorrect / total

        return {
            "score": score,
            "overall": {
                "substitutions": total_substitutions,
                "deletions": total_deletions,
                "insertions": total_insertions,
                "hits": total_hits,
            },
        }

    def _compute_single_pred_multi_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        **kwargs
    ):
        if "concatenate_texts" in kwargs:
            warnings.warn("Option 'concatenate_texts' is only available in single-pred & single-ref setting.")

        total_substitutions = 0
        total_deletions = 0
        total_insertions = 0
        total_hits = 0
        total_refs = 0
        scores = []
        for pred, refs in zip(predictions, references):
            pred_scores = []
            for ref in refs:
                score, substitutions, deletions, insertions, hits = self._compute_wer_score(
                    predictions=pred, references=ref
                )
                pred_scores.append(score)
                total_substitutions += substitutions
                total_deletions += deletions
                total_insertions += insertions
                total_hits += hits
                total_refs += 1
            reduced_score = float(reduce_fn(pred_scores))
            scores.append(reduced_score)

        return {
            "score": float(reduce_fn(scores)),
            "overall": {
                "substitutions": total_substitutions / total_refs,
                "deletions": total_deletions / total_refs,
                "insertions": total_insertions / total_refs,
                "hits": total_hits / total_refs,
            },
        }

    def _compute_multi_pred_multi_ref(
        self,
        predictions: LanguageGenerationInstance,
        references: LanguageGenerationInstance,
        reduce_fn: Callable = None,
        **kwargs
    ):
        if "concatenate_texts" in kwargs:
            warnings.warn("Option 'concatenate_texts' is only available in single-pred & single-ref setting.")

        total_substitutions = 0
        total_deletions = 0
        total_insertions = 0
        total_hits = 0
        total_refs = 0
        scores = []
        for preds, refs in zip(predictions, references):
            pred_scores = []
            for pred in preds:
                for ref in refs:
                    score, substitutions, deletions, insertions, hits = self._compute_wer_score(
                        predictions=pred, references=ref
                    )
                    pred_scores.append(score)
                    total_substitutions += substitutions
                    total_deletions += deletions
                    total_insertions += insertions
                    total_hits += hits
                    total_refs += 1
            reduced_score = float(reduce_fn(pred_scores))
            scores.append(reduced_score)

        return {
            "score": float(reduce_fn(scores)),
            "overall": {
                "substitutions": total_substitutions / total_refs,
                "deletions": total_deletions / total_refs,
                "insertions": total_insertions / total_refs,
                "hits": total_hits / total_refs,
            },
        }



import json
import os
import pickle
from typing import Any, Dict, Union


def json_load(fp: str) -> Union[Dict, None]:
    """Try loading json and return parsed dictionary. Returns None if file does not exists,
    or in case of a serialization error."""
    try:
        with open(fp, "r") as jf:
            _json = json.load(jf)
    except FileNotFoundError:
        return
    except json.JSONDecodeError:
        return
    else:
        return _json


def json_save(obj: Dict, fp: str, overwrite: bool = True) -> None:
    """Saves a dictionary as json file to given fp."""
    if os.path.exists(fp) and not overwrite:
        raise ValueError(f"Path {fp} already exists. To overwrite, use `overwrite=True`.")

    with open(fp, "w") as jf:
        json.dump(obj, jf)


def pickle_load(fp: str) -> Any:
    try:
        with open(fp, "rb") as pkl:
            _obj = pickle.load(pkl)
    except FileNotFoundError:
        return
    else:
        return _obj


def pickle_save(obj: Dict, fp: str, overwrite: bool = True) -> None:
    """Saves a dictionary as json file to given fp."""
    if os.path.exists(fp) and not overwrite:
        raise ValueError(f"Path {fp} already exists. To overwrite, use overwrite=True.")

    with open(fp, "wb") as pkl:
        pickle.dump(obj, pkl)



from jury.metrics._core import MetricAlias
from jury.metrics.prism.prism_for_language_generation import PrismForLanguageGeneration

__main_class__ = "Prism"


class Prism(MetricAlias):
    _SUBCLASS = PrismForLanguageGeneration



# coding=utf-8
# Copyright 2021 Open Business Software Solutions, The HuggingFace Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
SacreBLEU metric. The part of this file is adapted from SacreBLEU implementation
of datasets package. See
https://github.com/huggingface/datasets/blob/master/metrics/sacrebleu/sacrebleu.py
"""

import math
from typing import Callable, Dict, Sequence

import datasets
from packaging import version

from jury.collator import Collator
from jury.metrics._core import MetricForLanguageGeneration
from jury.metrics._core.utils import PackagePlaceholder, get_token_lengths, requirement_message
from jury.tokenizer import DefaultTokenizer

# `import sacrebleu as scb` placeholder
scb = PackagePlaceholder(version="2.0.0")

__class_names__ = {"sacrebleu": "Sacrebleu"}


_CITATION = """\
@inproceedings{post-2018-call,
    title = "A Call for Clarity in Reporting {BLEU} Scores",
    author = "Post, Matt",
    booktitle = "Proceedings of the Third Conference on Machine Translation: Research Papers",
    month = oct,
    year = "2018",
    address = "Belgium, Brussels",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/W18-6319",
    pages = "186--191",
}
"""

_DESCRIPTION = """\
SacreBLEU provides hassle-free computation of shareable, comparable, and reproducible BLEU scores.
Inspired by Rico Sennrich's `multi-bleu-detok.perl`, it produces the official WMT scores but works with plain text.
It also knows all the standard test sets and handles downloading, processing, and tokenization for you.

See the [README.md] file at https://github.com/mjpost/sacreBLEU for more information.
"""

_KWARGS_DESCRIPTION = """
Produces BLEU scores along with its sufficient statistics
from a source against one or more references.

Args:
    predictions: list of predictions to score. Each predictions
        should be a string with tokens separated by spaces.
    references: list of reference for each prediction. Each
        reference should be a string with tokens separated by spaces.
    smooth_method: The smoothing method to use. (Default: 'exp').
    smooth_value: The smoothing value. Only valid for 'floor' and 'add-k'. (Defaults: floor: 0.1, add-k: 1).
    tokenize: Tokenization method to use for BLEU. If not provided, defaults to 'zh' for Chinese, 'ja-mecab' for
        Japanese and '13a' (mteval) otherwise.
    lowercase: Lowercase the data. If True, enables case-insensitivity. (Default: False).
    force: Insist that your tokenized input is actually detokenized.

Returns:
    'score': BLEU score,
    'counts': Counts,
    'totals': Totals,
    'precisions': Precisions,
    'bp': Brevity penalty,
    'sys_len': predictions length,
    'ref_len': reference length,
    'adjusted_precision': adjusted precisions with corrections for multiple predictions cases.

Examples:

    >>> sacrebleu = jury.load_metric("sacrebleu")
    >>> predictions = [["the cat is on the mat", "There is cat playing on the mat"], ["Look! a wonderful day."]]
    >>> references = [
        ["the cat is playing on the mat.", "The cat plays on the mat."], 
        ["Today is a wonderful day", "The weather outside is wonderful."]
    ]
    >>> results = sacrebleu.compute(predictions=predictions, references=references)
    >>> print(results)
    {
      "sacrebleu": {
        "score": 0.32377227131456443,
        "counts": [
          11,
          6,
          3,
          0
        ],
        "totals": [
          13,
          11,
          9,
          7
        ],
        "precisions": [
          0.8461538461538461,
          0.5454545454545454,
          0.33333333333333337,
          0.07142857142857144
        ],
        "bp": 1.0,
        "sys_len": 11,
        "ref_len": 12,
        "adjusted_precisions": [
          0.8461538461538461,
          0.5454545454545454,
          0.33333333333333337,
          0.07142857142857144
        ]
      }
    }
"""


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class SacrebleuForLanguageGeneration(MetricForLanguageGeneration):
    def _download_and_prepare(self, dl_manager):
        global scb
        try:
            import sacrebleu as scb
        except ModuleNotFoundError:
            raise ModuleNotFoundError(requirement_message(path="Sacrebleu", package_name="sacrebleu"))
        else:
            super(SacrebleuForLanguageGeneration, self)._download_and_prepare(dl_manager)

    def _info(self):
        if version.parse(scb.__version__) < version.parse("1.4.12"):
            raise ImportWarning(
                "To use `sacrebleu`, the module `sacrebleu>=1.4.12` is required, and the current version of `sacrebleu` doesn't match this condition.\n"
                'You can install it with `pip install "sacrebleu>=1.4.12"`.'
            )
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            homepage="https://github.com/mjpost/sacreBLEU",
            inputs_description=_KWARGS_DESCRIPTION,
            features=self._default_features,
            codebase_urls=["https://github.com/mjpost/sacreBLEU"],
            reference_urls=[
                "https://github.com/mjpost/sacreBLEU",
                "https://en.wikipedia.org/wiki/BLEU",
                "https://towardsdatascience.com/evaluating-text-output-in-nlp-bleu-at-your-own-risk-e8609665a213",
            ],
        )

    def _tokenize(self, seq: Sequence[str]) -> Sequence[Sequence[str]]:
        tokenizer = DefaultTokenizer()
        return [tokenizer.tokenize(s) for s in seq]

    def _validate_references(self, references: Collator) -> None:
        references_per_prediction = len(references[0])
        if any(len(refs) != references_per_prediction for refs in references):
            raise ValueError("Sacrebleu requires the same number of references for each prediction")

    def _compute_single_pred_single_ref(
        self,
        predictions: Collator,
        references: Collator,
        reduce_fn: Callable = None,
        smooth_method="exp",
        smooth_value=None,
        force=False,
        lowercase=False,
        tokenize=None,
        use_effective_order=False,
    ):
        output = scb.corpus_bleu(
            predictions,
            references,
            smooth_method=smooth_method,
            smooth_value=smooth_value,
            force=force,
            lowercase=lowercase,
            use_effective_order=use_effective_order,
            **(dict(tokenize=tokenize) if tokenize else {}),
        )
        output_dict = {
            "score": output.score / 100,
            "counts": output.counts,
            "totals": output.totals,
            "precisions": [p / 100 for p in output.precisions],
            "bp": output.bp,
            "sys_len": output.sys_len,
            "ref_len": output.ref_len,
        }
        return output_dict

    def _compute_single_pred_multi_ref(
        self,
        predictions: Collator,
        references: Collator,
        reduce_fn: Callable = None,
        smooth_method="exp",
        smooth_value=None,
        force=False,
        lowercase=False,
        tokenize=None,
        use_effective_order=False,
    ):
        # SacreBleu inherently supports multiple references.
        return self._compute_single_pred_single_ref(
            predictions=predictions,
            references=references,
            reduce_fn=reduce_fn,
            smooth_method=smooth_method,
            smooth_value=smooth_value,
            force=force,
            lowercase=lowercase,
            tokenize=tokenize,
            use_effective_order=use_effective_order,
        )

    def _compute_multi_pred_multi_ref(
        self,
        predictions: Collator,
        references: Collator,
        reduce_fn: Callable = None,
        smooth_method="exp",
        smooth_value=None,
        force=False,
        lowercase=False,
        tokenize=None,
        use_effective_order=False,
    ):
        flattened_predictions = []
        matched_references = []
        adjusted_prediction_length = 0
        for preds, refs in zip(predictions, references):
            n_preds = len(preds)
            tokenized_preds = self._tokenize(preds)
            adjusted_prediction_length += get_token_lengths(tokenized_preds, reduce_fn=max)
            flattened_predictions.extend([pred for pred in preds])
            matched_references.extend([refs] * n_preds)
        flattened_predictions = Collator(flattened_predictions, keep=True)
        matched_references = Collator(matched_references, keep=True)
        score = self._compute_single_pred_multi_ref(
            predictions=flattened_predictions,
            references=matched_references,
            reduce_fn=reduce_fn,
            smooth_method=smooth_method,
            smooth_value=smooth_value,
            force=force,
            lowercase=lowercase,
            tokenize=tokenize,
            use_effective_order=use_effective_order,
        )
        prediction_length, reference_length = score["sys_len"], score["ref_len"]
        ratio = prediction_length / reference_length
        adjusted_ratio = adjusted_prediction_length / reference_length
        if ratio > 1.0:
            adjusted_bp = 1.0
            scb_score = score["score"]
        else:
            bp = math.exp(1 - 1.0 / ratio)
            adjusted_bp = math.exp(1 - 1.0 / adjusted_ratio)
            scb_score = score["score"] * (adjusted_bp / bp)

        scb_score = scb_score * n_preds
        precisions = [p * n_preds for p in score["precisions"]]

        score.update(
            {
                "score": scb_score,
                "adjusted_precisions": precisions,
                "bp": adjusted_bp,
                "sys_len": adjusted_prediction_length,
            }
        )

        return score

    def evaluate(
        self, predictions: Collator, references: Collator, reduce_fn: Callable = None, **kwargs
    ) -> Dict[str, float]:
        if predictions.can_collapse() and references.can_collapse():
            predictions = predictions.collapse()
            eval_fn = self._compute_single_pred_single_ref
        elif predictions.can_collapse() and not references.can_collapse():
            predictions = predictions.collapse()
            eval_fn = self._compute_single_pred_multi_ref
        else:
            eval_fn = self._compute_multi_pred_multi_ref
        self._validate_references(references)
        return eval_fn(predictions=predictions, references=references, reduce_fn=reduce_fn, **kwargs)
